from contextlib import contextmanager
import datetime
import json
import random
import re
import sys
import time
import eventlet
eventlet.sleep()
eventlet.monkey_patch()
import docassemble.base.config  # noqa: E402
docassemble.base.config.load(arguments=sys.argv)
from docassemble.base.config import daconfig  # noqa: E402
import docassemble.base.functions  # noqa: E402
import docassemble.base.util  # noqa: E402
docassemble.base.functions.server_context.context = 'websockets'
from docassemble.base.functions import get_default_timezone, word  # noqa: E402,F401 # pylint: disable=unused-import
from docassemble.base.logger import logmessage  # noqa: E402
import docassemble.webapp.daredis  # noqa: E402
from docassemble.webapp.app_socket import app, db, socketio  # noqa: E402
from docassemble.webapp.backend import nice_utc_date, fetch_user_dict, get_chat_log, encrypt_phrase, pack_phrase, fix_pickle_obj, get_session  # noqa: E402
from docassemble.webapp.daredis import r as rr, redis_host, redis_port, redis_offset  # noqa: E402
from docassemble.webapp.users.models import UserModel, ChatLog  # noqa: E402
from docassemblekvsession import KVSessionExtension  # noqa: E402
from flask import session, request  # noqa: E402
from flask_socketio import join_room  # noqa: E402
from simplekv.memory.redisstore import RedisStore  # noqa: E402
from sqlalchemy import select  # noqa: E402
from sqlalchemy.orm import sessionmaker, joinedload  # noqa: E402
import netifaces as ni  # noqa: E402 # pylint: disable=import-error
import redis  # noqa: E402

store = RedisStore(docassemble.webapp.daredis.r_store)
kv_session = KVSessionExtension(store, app)
threads = {}
secrets = {}
Session = sessionmaker(bind=db)


@contextmanager
def session_scope():
    """Provide a transactional scope around a series of operations."""
    dbsession = Session()
    db.session = dbsession
    try:
        yield dbsession
        dbsession.commit()
    except:
        dbsession.rollback()
        raise
    finally:
        dbsession.close()


def obtain_lock(user_code, filename):
    key = 'da:lock:' + user_code + ':' + filename
    found = False
    count = 4
    while count > 0:
        record = rr.get(key)
        if record:
            logmessage("obtain_lock: waiting for " + key)
            time.sleep(1.0)
        else:
            found = False
            break
        found = True
        count -= 1
    if found:
        logmessage("Request for " + key + " deadlocked")
        release_lock(user_code, filename)
    pipe = rr.pipeline()
    pipe.set(key, 1)
    pipe.expire(key, 4)
    pipe.execute()


def release_lock(user_code, filename):
    key = 'da:lock:' + user_code + ':' + filename
    rr.delete(key)

# @app.teardown_appcontext
# def close_db(error):
#     # logmessage("Teardown of app context")
#     if hasattr(db, 'engine'):
#         # logmessage("Tearing down")
#         db.engine.dispose()


def background_thread(sid=None, user_id=None, temp_user_id=None):
    if user_id is not None:
        user_id = int(user_id)
    if temp_user_id is not None:
        temp_user_id = int(temp_user_id)
    logmessage("Started client thread for " + str(sid) + " who is " + str(user_id) + " or " + str(temp_user_id))
    if user_id is None:
        person = None
        user_is_temp = True
    else:
        with session_scope() as dbsession:
            person = dbsession.execute(select(UserModel).options(joinedload(UserModel.roles)).filter_by(id=user_id)).scalar()
        user_is_temp = False
    if not (person is not None and person.timezone is not None):
        r = redis.StrictRedis(host=redis_host, port=redis_port, db=redis_offset)

    partners = set()
    pubsub = r.pubsub()
    pubsub.subscribe([sid])
    for item in pubsub.listen():
        with session_scope() as dbsession:
            logmessage("0\n" + repr(item))
            if item['type'] != 'message':
                continue
            # logmessage("sid: " + str(sid) + ":")
            data = None
            try:
                data = json.loads(item['data'].decode())
            except:
                logmessage("  JSON parse error: " + str(item['data'].decode()))
                continue
            if data.get('message', None) == "KILL" and (('sid' in data and data['sid'] == sid) or 'sid' not in data):
                pubsub.unsubscribe(sid)
                logmessage("  interview unsubscribed and finished for " + str(sid))
                break
            logmessage("  Got something for sid " + str(sid) + " from " + data.get('origin', "Unknown origin"))
            if 'messagetype' in data:
                if data['messagetype'] == 'chat':
                    # logmessage("  Emitting interview chat message: " + str(data['message']['message']))
                    data['message']['is_self'] = bool((user_is_temp is True and str(temp_user_id) == str(data['message'].get('temp_user_id', None))) or (user_is_temp is False and str(user_id) == str(data['message'].get('user_id', None))))
                    socketio.emit('chatmessage', {'i': data['yaml_filename'], 'uid': data['uid'], 'userid': data['user_id'], 'data': data['message']}, namespace='/wsinterview', room=sid)
                elif data['messagetype'] == 'chatready':
                    pubsub.subscribe(data['sid'])
                    partners.add(data['sid'])
                    logmessage("chatready 2")
                    socketio.emit('chatready', {}, namespace='/wsinterview', room=sid)
                elif data['messagetype'] == 'departure':
                    if data['sid'] in partners:
                        partners.remove(data['sid'])
                    socketio.emit('departure', {'numpartners': len(partners)}, namespace='/wsinterview', room=sid)
                elif data['messagetype'] == 'block':
                    if data['sid'] in partners:
                        partners.remove(data['sid'])
                    socketio.emit('departure', {'numpartners': len(partners)}, namespace='/wsinterview', room=sid)
                elif data['messagetype'] == 'chatpartner':
                    partners.add(data['sid'])
                elif data['messagetype'] == 'controllerchanges':
                    socketio.emit('controllerchanges', {'parameters': data['parameters'], 'clicked': data['clicked']}, namespace='/wsinterview', room=sid)
                elif data['messagetype'] == 'controllerstart':
                    socketio.emit('controllerstart', {}, namespace='/wsinterview', room=sid)
                elif data['messagetype'] == 'controllerexit':
                    socketio.emit('controllerexit', {}, namespace='/wsinterview', room=sid)
                # elif data['messagetype'] == "newpage":
                #     logmessage("  Got new page for interview")
                #     try:
                #         obj = json.loads(r.get(data['key']))
                #     except:
                #         logmessage("  newpage JSON parse error")
                #         continue
                #     socketio.emit('newpage', {'obj': obj}, namespace='/wsinterview', room=sid)
            logmessage('  exiting interview thread for sid ' + str(sid))


@socketio.on('start_being_controlled', namespace='/wsinterview')
def interview_start_being_controlled(data):
    # logmessage("received start_being_controlled")
    yaml_filename = data['i']
    session_info = get_session(yaml_filename)
    if session_info is not None:
        session_id = session_info['uid']
        the_user_id = session.get('user_id', 't' + str(session.get('tempuser', None)))
        key = 'da:input:uid:' + str(session_id) + ':i:' + str(yaml_filename) + ':userid:' + str(the_user_id)
        rr.publish(key, json.dumps(dict(message='start_being_controlled', key=re.sub(r'^da:input:uid:', 'da:session:uid:', key))))


@socketio.on('chat_log', namespace='/wsinterview')
def chat_log(data):
    with session_scope():
        yaml_filename = data['i']
        user_dict = get_dict(yaml_filename)
        if user_dict is None:
            return
        chat_mode = user_dict['_internal']['livehelp']['mode']
        session_info = get_session(yaml_filename)
        if session_info is not None:
            session_id = session_info['uid']
            user_id = session.get('user_id', None)
            if user_id is None:
                temp_user_id = session.get('tempuser', None)
            else:
                temp_user_id = None
            if user_id is not None:
                user_id = int(user_id)
            if temp_user_id is not None:
                temp_user_id = int(temp_user_id)
            secret = request.cookies.get('secret', None)
            if secret is not None:
                secret = str(secret)
            # logmessage("chat_log: " + str(repr(user_id)) + " " + str(repr(temp_user_id)))
            messages = get_chat_log(chat_mode, yaml_filename, session_id, user_id, temp_user_id, secret, user_id, temp_user_id)
            socketio.emit('chat_log', {'data': messages}, namespace='/wsinterview', room=request.sid)
    # logmessage("Interview: sending back " + str(len(messages)) + " messages")


@socketio.on('transmit', namespace='/wsinterview')
def handle_message(message):
    session_info = get_session(message['i'])
    if session_info is not None:
        rr.publish(session_info['uid'], json.dumps(dict(origin='client', room=request.sid, message=message['data'])))


@socketio.on('terminate', namespace='/wsinterview')
def terminate_interview_connection():
    logmessage("terminate_interview_connection")
    # hopefully the disconnect will be triggered
    # if request.sid in threads:
    #     rr.publish(request.sid, json.dumps(dict(origin='client', message='KILL', sid=request.sid)))
    socketio.emit('terminate', {}, namespace='/wsinterview', room=request.sid)
    # disconnect()


@socketio.on('chatmessage', namespace='/wsinterview')
def chat_message(data):
    nowtime = datetime.datetime.utcnow()
    yaml_filename = data['i']
    session_info = get_session(yaml_filename)
    if session_info is not None:
        session_id = session_info['uid']
        encrypted = session_info['encrypted']
    else:
        session_id = None
        encrypted = True
    secret = request.cookies.get('secret', None)
    if secret is not None:
        secret = str(secret)
    if encrypted:
        message = encrypt_phrase(data['data'], secret)
    else:
        message = pack_phrase(data['data'])
    user_id = session.get('user_id', None)
    if user_id is None:
        temp_user_id = session.get('tempuser', None)
    else:
        temp_user_id = None
    if user_id is not None:
        user_id = int(user_id)
    if temp_user_id is not None:
        temp_user_id = int(temp_user_id)
    with session_scope() as dbsession:
        user_dict = get_dict(yaml_filename)
        chat_mode = user_dict['_internal']['livehelp']['mode']
        open_to_peer = bool(chat_mode in ['peer', 'peerhelp'])
        record = ChatLog(filename=yaml_filename, key=session_id, message=message, encrypted=encrypted, modtime=nowtime, temp_user_id=temp_user_id, user_id=user_id, open_to_peer=open_to_peer, temp_owner_id=temp_user_id, owner_id=user_id)
        dbsession.add(record)
        dbsession.commit()
        if user_id is not None:
            person = dbsession.execute(select(UserModel).options(joinedload(UserModel.roles)).filter_by(id=user_id)).scalar()
        else:
            person = None
        modtime = nice_utc_date(nowtime)
        if person is None:
            rr.publish(request.sid, json.dumps(dict(origin='client', messagetype='chat', sid=request.sid, yaml_filename=yaml_filename, uid=session_id, user_id='t' + str(temp_user_id), message=dict(id=record.id, temp_user_id=record.temp_user_id, modtime=modtime, message=data['data'], roles=['user'], mode=chat_mode))))
        else:
            role_list = [role.name for role in person.roles]
            if len(role_list) == 0:
                role_list = ['user']
            rr.publish(request.sid, json.dumps(dict(origin='client', messagetype='chat', sid=request.sid, yaml_filename=yaml_filename, uid=session_id, user_id=user_id, message=dict(id=record.id, user_id=record.user_id, first_name=person.first_name, last_name=person.last_name, email=person.email, modtime=modtime, message=data['data'], roles=role_list, mode=chat_mode))))
    # logmessage('received chat message from sid ' + str(request.sid) + ': ' + data['data'])


def wait_for_channel(the_rr, channel):
    times = 0
    while times < 5 and the_rr.publish(channel, json.dumps(dict(messagetype='ping'))) == 0:
        times += 1
        time.sleep(0.5)
    if times >= 5:
        return False
    return True


@socketio.on('connect', namespace='/wsinterview')
def on_interview_connect():
    with session_scope():
        logmessage("Client connected on interview")
        join_room(request.sid)
        interview_connect(request.args['i'])
        rr.publish('da:monitor', json.dumps(dict(messagetype='refreshsessions')))


@socketio.on('connectagain', namespace='/wsinterview')
def on_interview_reconnect(data):
    with session_scope():
        logmessage("Client reconnected on interview")
        interview_connect(data['i'])
        rr.publish('da:monitor', json.dumps(dict(messagetype='refreshsessions')))
        socketio.emit('reconnected', {}, namespace='/wsinterview', room=request.sid)


def interview_connect(yaml_filename):
    session_info = get_session(yaml_filename)
    if session_info is not None:
        session_id = session_info['uid']
        user_dict, is_encrypted = get_dict_encrypt(yaml_filename)
        if is_encrypted:
            secret = request.cookies.get('secret', None)
        else:
            secret = None
        if secret is not None:
            secret = str(secret)
        if user_dict is None:
            logmessage("user_dict did not exist.")
            socketio.emit('terminate', {}, namespace='/wsinterview', room=request.sid)
            return

        chat_info = user_dict['_internal']['livehelp']
        if chat_info['availability'] == 'unavailable':
            logmessage("Socket started but chat is unavailable.")
            socketio.emit('terminate', {}, namespace='/wsinterview', room=request.sid)
            return
        # logmessage('chat info is ' + str(chat_info))
        peer_ok = bool(user_dict['_internal']['livehelp']['mode'] in ['peer', 'peerhelp'])
        the_user_id = session.get('user_id', 't' + str(session.get('tempuser', None)))

        if request.sid not in threads:
            # logmessage('Starting thread for sid ' + str(request.sid))
            threads[request.sid] = socketio.start_background_task(target=background_thread, sid=request.sid, user_id=session.get('user_id', None), temp_user_id=session.get('tempuser', None))
        channel_up = wait_for_channel(rr, request.sid)
        if not channel_up:
            logmessage("Channel did not come up.")
            socketio.emit('terminate', {}, namespace='/wsinterview', room=request.sid)
            return
        lkey = 'da:ready:uid:' + str(session_id) + ':i:' + str(yaml_filename) + ':userid:' + str(the_user_id)
        # logmessage("Searching: " + lkey)
        lkey_exists = bool(rr.exists(lkey))
        if lkey_exists is False and peer_ok is False:
            logmessage("Key does not exist: " + lkey + ".")
            # socketio.emit('terminate', {}, namespace='/wsinterview', room=request.sid)
            # return
        failed_to_find_partner = True
        found_help = False
        if lkey_exists:
            partner_keys = rr.lrange(lkey, 0, -1)
            # logmessage("partner_keys is: " + str(type(partner_keys)) + " " + str(partner_keys))
            if partner_keys is None and not peer_ok:
                logmessage("No partner keys: " + lkey + ".")
                socketio.emit('terminate', {}, namespace='/wsinterview', room=request.sid)
                return
            rr.delete(lkey)
            for pkey in partner_keys:
                pkey = pkey.decode()
                # logmessage("Considering: " + pkey)
                partner_sid = rr.get(pkey)
                if partner_sid is not None:
                    partner_sid = partner_sid.decode()
                    is_help = bool(re.match(r'^da:monitor:available:.*', pkey))
                    if is_help and found_help:
                        continue
                    # logmessage("Trying to pub to " + str(partner_sid) + " from " + str(pkey))
                    listeners = rr.publish(partner_sid, json.dumps(dict(messagetype='chatready', uid=session_id, i=yaml_filename, userid=the_user_id, secret=secret, sid=request.sid)))
                    # logmessage("Listeners: " + str(listeners))
                    if re.match(r'^da:interviewsession.*', pkey):
                        rr.publish(request.sid, json.dumps(dict(messagetype='chatready', sid=partner_sid)))
                    else:
                        rr.publish(request.sid, json.dumps(dict(messagetype='chatpartner', sid=partner_sid)))
                    if listeners > 0:
                        if is_help:
                            found_help = True
                        failed_to_find_partner = False
        if failed_to_find_partner and peer_ok is False:
            logmessage("Unable to reach any potential chat partners.")
            # socketio.emit('terminate', {}, namespace='/wsinterview', room=request.sid)
            # return
        key = 'da:interviewsession:uid:' + str(session_id) + ':i:' + str(yaml_filename) + ':userid:' + str(the_user_id)
        rr.set(key, request.sid)


@socketio.on('manual_disconnect', namespace='/wsinterview')
def on_interview_manual_disconnect(data):
    logmessage("Client manual disconnect")
    yaml_filename = data['i']
    session_info = get_session(yaml_filename)
    session_id = session_info['uid']
    the_user_id = session.get('user_id', 't' + str(session.get('tempuser', None)))
    if request.sid in secrets:
        del secrets[request.sid]
    if session_id is not None:
        rr.delete('da:interviewsession:uid:' + str(session_id) + ':i:' + str(yaml_filename) + ':userid:' + str(the_user_id))
        key = 'da:session:uid:' + str(session_id) + ':i:' + str(yaml_filename) + ':userid:' + str(the_user_id)
        rr.expire(key, 10)
        rr.publish(request.sid, json.dumps(dict(origin='client', message='KILL', sid=request.sid)))


@socketio.on('disconnect', namespace='/wsinterview')
def on_interview_disconnect():
    logmessage('Client disconnected from interview')


def get_current_info(yaml_filename, session_id, secret):
    url_root = daconfig.get('url root', 'http://localhost') + daconfig.get('root', '/')
    url = url_root + 'interview'
    return dict(user=dict(is_anonymous=False, is_authenticated=True, email='admin@admin.com', theid=1, the_user_id=1, roles=['admin'], firstname='Admin', lastname='User', nickname='', country='', subdivisionfirst='', subdivisionsecond='', subdivisionthird='', organization='', location=None, session_uid='admin', device_id='admin'), session=session_id, secret=secret, yaml_filename=yaml_filename, url=url, url_root=url_root, encrypted=True, action=None, interface='chat', arguments={})


def get_dict(yaml_filename):
    session_info = get_session(yaml_filename)
    if session_info is not None:
        session_id = session_info['uid']
    else:
        session_id = None
    secret = request.cookies.get('secret', None)
    if secret is not None:
        secret = str(secret)
    if session_id is None or yaml_filename is None:
        logmessage('Attempt to get dictionary where session not defined')
        return None
    # obtain_lock(session_id, yaml_filename)
    docassemble.base.functions.this_thread.current_info = get_current_info(yaml_filename, session_id, secret)
    try:
        steps, user_dict, is_encrypted = fetch_user_dict(session_id, yaml_filename, secret=secret)  # pylint: disable=unused-variable
    except Exception as err:
        # release_lock(session_id, yaml_filename)
        logmessage('get_dict: attempt to get dictionary failed: ' + str(err))
        return None
    docassemble.base.functions.this_thread.current_info['encrypted'] = is_encrypted
    # release_lock(session_id, yaml_filename)
    return user_dict


def get_dict_encrypt(yaml_filename):
    session_info = get_session(yaml_filename)
    if session_info is not None:
        session_id = session_info['uid']
    else:
        session_id = None
    secret = request.cookies.get('secret', None)
    if secret is not None:
        secret = str(secret)
    if session_id is None or yaml_filename is None:
        logmessage('Attempt to get dictionary where session not defined')
        return None, None
    # obtain_lock(session_id, yaml_filename)
    docassemble.base.functions.this_thread.current_info = get_current_info(yaml_filename, session_id, secret)
    try:
        steps, user_dict, is_encrypted = fetch_user_dict(session_id, yaml_filename, secret=secret)  # pylint: disable=unused-variable
    except Exception as err:
        # release_lock(session_id, yaml_filename)
        logmessage('get_dict_encrypt: attempt to get dictionary failed: ' + str(err))
        return None, None
    docassemble.base.functions.this_thread.current_info['encrypted'] = is_encrypted
    # release_lock(session_id, yaml_filename)
    return user_dict, is_encrypted

# monitor


def monitor_thread(sid=None, user_id=None):
    logmessage("Started monitor thread for " + str(sid) + " who is " + str(user_id))
    with session_scope() as dbsession:
        if user_id is not None:
            person = dbsession.execute(select(UserModel).options(joinedload(UserModel.roles)).filter_by(id=user_id)).scalar()
        else:
            person = None
    r = redis.StrictRedis(host=redis_host, port=redis_port, db=redis_offset)
    listening_sids = set()
    pubsub = r.pubsub()
    pubsub.subscribe(['da:monitor', sid])
    for item in pubsub.listen():
        with session_scope() as dbsession:
            logmessage("1\n" + repr(item))
            if item['type'] != 'message':
                continue
            # logmessage("monitor sid: " + str(sid) + ":")
            data = None
            try:
                data = json.loads(item['data'].decode())
            except:
                logmessage("  monitor JSON parse error: " + item['data'].decode())
                continue
            if 'message' in data and data['message'] == "KILL":
                if item['channel'].decode() == str(sid):
                    logmessage("  monitor unsubscribed from all")
                    pubsub.unsubscribe()
                    for interview_sid in listening_sids:
                        r.publish(interview_sid, json.dumps(dict(messagetype='departure', sid=sid)))
                    break
                if item['channel'].decode() != 'da:monitor':
                    pubsub.unsubscribe(item['channel'].decode())
                    if data['sid'] in listening_sids:
                        listening_sids.remove(data['sid'])
                    logmessage("  monitor unsubscribed from " + item['channel'].decode())
                continue
            logmessage("  Got something for monitor")
            if 'messagetype' in data:
                # if data['messagetype'] == 'abortcontroller':
                #    socketio.emit('abortcontroller', {'key': data['key']}, namespace='/monitor', room=sid)
                if data['messagetype'] == 'sessionupdate':
                    # logmessage("  Got a session update: " + str(data['session']))
                    # logmessage("  Got a session update")
                    socketio.emit('sessionupdate', {'key': data['key'], 'session': data['session']}, namespace='/monitor', room=sid)
                if data['messagetype'] == 'chatready':
                    pubsub.subscribe(data['sid'])
                    listening_sids.add(data['sid'])
                    secrets[data['sid']] = data['secret']
                    r.hset('da:monitor:chatpartners:' + str(user_id), 'da:interviewsession:uid:' + str(data['uid']) + ':i:' + str(data['i']) + ':userid:' + str(data['userid']), 1)
                    if str(data['userid']).startswith('t'):
                        name = word("anonymous visitor") + ' ' + str(data['userid'])[1:]
                    else:
                        person = dbsession.execute(select(UserModel).options(joinedload(UserModel.roles)).filter_by(id=data['userid'])).scalar()
                        if person.first_name:
                            name = str(person.first_name) + ' ' + str(person.last_name)
                        else:
                            name = str(person.email)
                    logmessage("chatready 1")
                    socketio.emit('chatready', {'uid': data['uid'], 'i': data['i'], 'userid': data['userid'], 'name': name}, namespace='/monitor', room=sid)
                if data['messagetype'] == 'block':
                    pubsub.unsubscribe(item['channel'].decode())
                    if item['channel'].decode() in listening_sids:
                        listening_sids.remove(item['channel'].decode())
                    logmessage("  monitor unsubscribed from " + item['channel'].decode())
                if data['messagetype'] == 'refreshsessions':
                    socketio.emit('refreshsessions', {}, namespace='/monitor', room=sid)
                if data['messagetype'] == 'chat':
                    # logmessage("  Emitting monitor chat message: " + str(data['message']['message']))
                    data['message']['is_self'] = bool(str(user_id) == str(data['message'].get('user_id', None)))
                    socketio.emit('chatmessage', {'i': data['yaml_filename'], 'uid': data['uid'], 'userid': data['user_id'], 'data': data['message']}, namespace='/monitor', room=sid)
                if data['messagetype'] == 'chatstop':
                    logmessage("  Chat termination for sid " + data['sid'])
                    pubsub.unsubscribe(data['sid'])
                    if data['sid'] in secrets:
                        del secrets[data['sid']]
                    r.hdel('da:monitor:chatpartners:' + str(user_id), 'da:interviewsession:uid:' + str(data['uid']) + ':i:' + str(data['i']) + ':userid:' + data['userid'])
                    socketio.emit('chatstop', {'uid': data['uid'], 'i': data['i'], 'userid': data['userid']}, namespace='/monitor', room=sid)
            logmessage('  exiting monitor thread for sid ' + str(sid))


@socketio.on('connect', namespace='/monitor')
def on_monitor_connect():
    if 'monitor' not in session:
        socketio.emit('terminate', {}, namespace='/monitor', room=request.sid)
        return
    logmessage('Client connected on monitor and will join room monitor')
    key = 'da:monitor:' + str(request.sid)
    pipe = rr.pipeline()
    pipe.set(key, 1)
    pipe.expire(key, 60)
    pipe.execute()
    join_room('monitor')
    join_room(request.sid)
    user_id = session.get('user_id', None)
    if request.sid not in threads:
        threads[request.sid] = socketio.start_background_task(target=monitor_thread, sid=request.sid, user_id=user_id)


@socketio.on('disconnect', namespace='/monitor')
def on_monitor_disconnect():
    user_id = session.get('user_id', None)
    logmessage('Client disconnected from monitor')
    rr.delete('da:monitor:' + str(request.sid))
    rr.expire('da:monitor:available:' + str(user_id), 5)
    for key in rr.keys('da:monitor:role:*:userid:' + str(user_id)):
        key = key.decode()
        rr.expire(key, 5)
    for key in rr.keys('da:phonecode:monitor:' + str(user_id) + ':uid:*'):
        key = key.decode()
        the_code = rr.get(key)
        if the_code is not None:
            the_code = the_code.decode()
            rr.expire('da:callforward:' + the_code, 5)
        rr.expire(key, 5)
    rr.expire('da:monitor:chatpartners:' + str(user_id), 5)
    rr.publish(request.sid, json.dumps(dict(message='KILL', sid=request.sid)))


@socketio.on('terminate', namespace='/monitor')
def terminate_monitor_connection():
    logmessage("terminate_monitor_connection")
    # hopefully the disconnect will be triggered
    # if request.sid in threads:
    #     rr.publish(request.sid, json.dumps(dict(origin='client', message='KILL', sid=request.sid)))
    socketio.emit('terminate', {}, namespace='/monitor', room=request.sid)
    # disconnect()


@socketio.on('block', namespace='/monitor')
def monitor_block(data):
    if 'monitor' not in session:
        socketio.emit('terminate', {}, namespace='/monitor', room=request.sid)
        return
    key = data.get('key', None)
    if key is None:
        logmessage("No key provided")
        return
    rr.set(re.sub(r'^da:session:', 'da:block:', key), 1)
    sid = rr.get(re.sub(r'^da:session:', 'da:interviewsession:', key))
    if sid is not None:
        sid = sid.decode()
        rr.publish(sid, json.dumps(dict(messagetype='block', sid=request.sid)))
        logmessage("Blocking")
    else:
        logmessage("Could not block because could not get sid")
    socketio.emit('block', {'key': key}, namespace='/monitor', room=request.sid)


@socketio.on('unblock', namespace='/monitor')
def monitor_unblock(data):
    if 'monitor' not in session:
        socketio.emit('terminate', {}, namespace='/monitor', room=request.sid)
        return
    key = data.get('key', None)
    if key is None:
        logmessage("No key provided")
        return
    logmessage("Unblocking")
    rr.delete(re.sub(r'^da:session:', 'da:block:', key))
    sid = rr.get(re.sub(r'^da:session:', 'da:interviewsession:', key))
    if sid is not None:
        sid = sid.decode()
        rr.publish(sid, json.dumps(dict(messagetype='chatpartner', sid=request.sid)))
    socketio.emit('unblock', {'key': key}, namespace='/monitor', room=request.sid)


def decode_dict(the_dict):
    out_dict = {}
    for k, v in the_dict.items():
        out_dict[k.decode()] = v.decode()
    return out_dict


@socketio.on('updatemonitor', namespace='/monitor')
def update_monitor(message):
    if 'monitor' not in session:
        socketio.emit('terminate', {}, namespace='/monitor', room=request.sid)
        return
    # logmessage('received message from ' + str(request.sid))
    available_for_chat = message['available_for_chat']
    new_subscribed_roles = message['subscribed_roles']
    new_phone_partners = message['phone_partners_to_add']
    term_phone_partners = message['phone_partners_to_terminate']
    phone_number = message['phone_number']
    phone_number_key = 'da:monitor:phonenumber:' + str(session['user_id'])
    if phone_number is None or phone_number == '':
        rr.delete(phone_number_key)
    else:
        pipe = rr.pipeline()
        pipe.set(phone_number_key, phone_number)
        pipe.expire(phone_number_key, 2592000)
        pipe.execute()
    phone_partners = {}
    prefix = 'da:phonecode:monitor:' + str(session['user_id']) + ':uid:'
    for key in term_phone_partners:
        key_for_phone_code = re.sub(r'da:session:uid:', prefix, key)
        the_code = rr.get(key_for_phone_code)
        if the_code is not None:
            the_code = the_code.decode()
            rr.delete(key_for_phone_code)
            rr.delete('da:callforward:' + the_code)
    if phone_number is None or phone_number == '':
        for key in rr.keys(prefix + '*'):
            key = key.decode()
            the_code = rr.get(key)
            if the_code is not None:
                the_code = the_code.decode()
                rr.delete(key)
                rr.delete('da:callforward:' + the_code)
    else:
        codes_in_use = set()
        for key in rr.keys('da:callforward:*'):
            key = key.decode()
            code = re.sub(r'^da:callforward:', '', key)
            codes_in_use.add(code)
        for key in rr.keys(prefix + '*'):
            key = key.decode()
            phone_partners[re.sub(r'^da:phonecode:monitor:[0-9]*:uid:', 'da:session:uid:', key)] = 1
        for key in new_phone_partners:
            if key in phone_partners:
                continue
            times = 0
            while times < 1000:
                times += 1
                code = "%04d" % random.randint(1000, 9999)
                if code in codes_in_use:
                    continue
                the_code = code
                new_key = re.sub(r'^da:session:uid:', prefix, key)
                code_key = 'da:callforward:' + str(code)
                pipe = rr.pipeline()
                pipe.set(new_key, code)
                pipe.set(code_key, phone_number)
                pipe.expire(new_key, 300)
                pipe.expire(code_key, 300)
                pipe.execute()
                phone_partners[key] = 1
                break
            if times >= 1000:
                logmessage("update_monitor: could not get a random integer")
    # logmessage('subscribed roles are type ' + str(type(new_subscribed_roles)) + " which is "  + str(new_subscribed_roles))
    monitor_key = 'da:monitor:' + str(request.sid)
    pipe = rr.pipeline()
    pipe.set(monitor_key, 1)
    pipe.expire(monitor_key, 60)
    pipe.execute()
    key = 'da:monitor:available:' + str(session['user_id'])
    key_exists = rr.exists(key)
    chat_partners = {}
    for cp_key in rr.hgetall('da:monitor:chatpartners:' + str(session['user_id'])):
        cp_key = cp_key.decode()
        if rr.get(cp_key) is None:
            rr.hdel('da:monitor:chatpartners:' + str(session['user_id']), cp_key)
        else:
            chat_partners[re.sub('^da:interviewsession:uid:', r'da:session:uid:', cp_key)] = 1
    # logmessage('daAvailableForChat is ' + str(available_for_chat) + " for key " + key)
    if available_for_chat:
        pipe = rr.pipeline()
        pipe.set(key, request.sid)
        pipe.expire(key, 60)
        pipe.execute()
    elif key_exists:
        # logmessage("Not available to chat; deleting da:monitor:role")
        pipe = rr.pipeline()
        pipe.delete(key)
        for avail_key in rr.keys('da:monitor:role:*:userid:' + str(session['user_id'])):
            pipe.delete(avail_key.decode())
        pipe.execute()
    avail_roles = []
    for key in rr.keys('da:chat:roletype:*'):
        avail_roles.append(re.sub(r'^da:chat:roletype:', r'', key.decode()))
    sub_role_key = 'da:monitor:userrole:' + str(session['user_id'])
    if rr.exists(sub_role_key):
        subscribed_roles = decode_dict(rr.hgetall(sub_role_key))
    else:
        subscribed_roles = {}
    del_mon_role_keys = []
    for role_key in list(new_subscribed_roles.keys()):
        if role_key not in avail_roles:
            # logmessage("role_key is " + str(role_key) + " which is " + str(type(role_key)))
            del new_subscribed_roles[role_key]
    for role_key in list(subscribed_roles.keys()):
        if role_key not in avail_roles:
            rr.hdel(sub_role_key, role_key)
            del_mon_role_keys.append('da:monitor:role:' + role_key + ':userid:' + str(session['user_id']))

    for role_key in list(new_subscribed_roles.keys()):
        if role_key not in subscribed_roles:
            rr.hset(sub_role_key, role_key, 1)
            subscribed_roles[role_key] = 1
    for role_key in list(subscribed_roles.keys()):
        if role_key not in new_subscribed_roles:
            rr.hdel(sub_role_key, role_key)
            del_mon_role_keys.append('da:monitor:role:' + role_key + ':userid:' + str(session['user_id']))
            del subscribed_roles[role_key]

    if len(del_mon_role_keys) > 0:
        pipe = rr.pipeline()
        for key in del_mon_role_keys:
            pipe.delete(key)
        pipe.execute()

    if available_for_chat and len(subscribed_roles) > 0:
        pipe = rr.pipeline()
        for role_key in list(subscribed_roles.keys()):
            key = 'da:monitor:role:' + role_key + ':userid:' + str(session['user_id'])
            pipe.set(key, 1)
            pipe.expire(key, 60)
        pipe.execute()
    keylist = []
    for key in rr.keys('da:session:*'):
        keylist.append(key.decode())
    sessions = {}
    for key in keylist:
        try:
            sessobj = fix_pickle_obj(rr.get(key))
        except:
            logmessage('error parsing value of ' + str(key) + " which was " + repr(rr.get(key)))
            continue
        if sessobj.get('chatstatus', None) != 'off':
            html = rr.get(re.sub(r'^da:session:', 'da:html:', key))
            if html is not None:
                html = html.decode()
                obj = json.loads(html)
                sessobj['browser_title'] = obj.get('browser_title', 'not available')
                sessobj['blocked'] = bool(rr.exists(re.sub(r'^da:session:', 'da:block:', key)))
                sessions[key] = sessobj
    socketio.emit('updatemonitor', {'available_for_chat': available_for_chat, 'subscribedRoles': subscribed_roles, 'sessions': sessions, 'availRoles': sorted(avail_roles), 'chatPartners': chat_partners, 'phonePartners': phone_partners}, namespace='/monitor', room=request.sid)


@socketio.on('chatmessage', namespace='/monitor')
def monitor_chat_message(data):
    if 'monitor' not in session:
        socketio.emit('terminate', {}, namespace='/monitor', room=request.sid)
        return
    key = data.get('key', None)
    # logmessage("Key is " + str(key))
    if key is None:
        logmessage("No key provided")
        return
    m = re.match(r'da:session:uid:(.*):i:(.*):userid:(.*)', key)
    if not m:
        logmessage("Invalid key provided")
        return
    session_id = m.group(1)
    yaml_filename = m.group(2)
    chat_user_id = m.group(3)
    key = 'da:interviewsession:uid:' + str(session_id) + ':i:' + str(yaml_filename) + ':userid:' + str(chat_user_id)
    sid = rr.get(key)
    if sid is None:
        logmessage("No sid for monitor chat message with key " + str(key))
        return
    sid = sid.decode()
    secret = secrets.get(sid, None)
    if secret is not None:
        secret = str(secret)
    # obtain_lock(session_id, yaml_filename)
    docassemble.base.functions.this_thread.current_info = get_current_info(yaml_filename, session_id, secret)
    with session_scope() as dbsession:
        try:
            steps, user_dict, encrypted = fetch_user_dict(session_id, yaml_filename, secret=secret)  # pylint: disable=unused-variable
        except Exception as err:
            # release_lock(session_id, yaml_filename)
            logmessage("monitor_chat_message: could not get dictionary: " + str(err))
            return
        # release_lock(session_id, yaml_filename)
        docassemble.base.functions.this_thread.current_info['encrypted'] = encrypted
        nowtime = datetime.datetime.utcnow()
        if encrypted:
            message = encrypt_phrase(data['data'], secret)
        else:
            message = pack_phrase(data['data'])
        user_id = session.get('user_id', None)
        if user_id is not None:
            user_id = int(user_id)
        person = dbsession.execute(select(UserModel).options(joinedload(UserModel.roles)).filter_by(id=user_id)).scalar()
        chat_mode = user_dict['_internal']['livehelp']['mode']
        m = re.match('t([0-9]+)', chat_user_id)
        if m:
            temp_owner_id = m.group(1)
            owner_id = None
        else:
            temp_owner_id = None
            owner_id = chat_user_id
        open_to_peer = bool(chat_mode in ['peer', 'peerhelp'])
        record = ChatLog(filename=yaml_filename, key=session_id, message=message, encrypted=encrypted, modtime=nowtime, user_id=user_id, temp_owner_id=temp_owner_id, owner_id=owner_id, open_to_peer=open_to_peer)
        dbsession.add(record)
        dbsession.commit()
        modtime = nice_utc_date(nowtime)
        role_list = [role.name for role in person.roles]
        if len(role_list) == 0:
            role_list = ['user']
        rr.publish(sid, json.dumps(dict(origin='client', messagetype='chat', sid=request.sid, yaml_filename=yaml_filename, uid=session_id, user_id=chat_user_id, message=dict(id=record.id, user_id=record.user_id, first_name=person.first_name, last_name=person.last_name, email=person.email, modtime=modtime, message=data['data'], roles=role_list, mode=chat_mode))))
    # logmessage('received chat message on monitor from sid ' + str(request.sid) + ': ' + data['data'])


@socketio.on('chat_log', namespace='/monitor')
def monitor_chat_log(data):
    with session_scope():
        if 'monitor' not in session:
            socketio.emit('terminate', {}, namespace='/monitor', room=request.sid)
            return
        key = data.get('key', None)
        scroll = data.get('scroll', True)
        # logmessage("Key is " + str(key))
        # logmessage("scroll is " + repr(scroll))
        if key is None:
            logmessage("No key provided")
            return
        m = re.match(r'da:session:uid:(.*):i:(.*):userid:(.*)', key)
        if not m:
            logmessage("Invalid key provided")
            return
        session_id = m.group(1)
        yaml_filename = m.group(2)
        chat_user_id = m.group(3)
        key = 'da:interviewsession:uid:' + str(session_id) + ':i:' + str(yaml_filename) + ':userid:' + str(chat_user_id)
        sid = rr.get(key)
        if sid is None:
            logmessage("No sid for monitor chat message with key " + str(key))
            return
        sid = sid.decode()
        secret = secrets.get(sid, None)
        if secret is not None:
            secret = str(secret)
        # obtain_lock(session_id, yaml_filename)
        docassemble.base.functions.this_thread.current_info = get_current_info(yaml_filename, session_id, secret)
        try:
            steps, user_dict, encrypted = fetch_user_dict(session_id, yaml_filename, secret=secret)  # pylint: disable=unused-variable
        except Exception as err:
            # release_lock(session_id, yaml_filename)
            logmessage("monitor_chat_log: could not get dictionary: " + str(err))
            return
        docassemble.base.functions.this_thread.current_info['encrypted'] = encrypted
        # release_lock(session_id, yaml_filename)
        chat_mode = user_dict['_internal']['livehelp']['mode']
        m = re.match('t([0-9]+)', chat_user_id)
        if m:
            temp_user_id = m.group(1)
            user_id = None
        else:
            temp_user_id = None
            user_id = chat_user_id
        self_user_id = session.get('user_id', None)
        if user_id is not None:
            user_id = int(user_id)
        if temp_user_id is not None:
            temp_user_id = int(temp_user_id)
        if self_user_id is not None:
            self_user_id = int(self_user_id)
        messages = get_chat_log(chat_mode, yaml_filename, session_id, user_id, temp_user_id, secret, self_user_id, None)
        socketio.emit('chat_log', {'uid': session_id, 'i': yaml_filename, 'userid': chat_user_id, 'mode': chat_mode, 'data': messages, 'scroll': scroll}, namespace='/monitor', room=request.sid)
        # logmessage("Monitor: sending back " + str(len(messages)) + " messages")

# observer


def observer_thread(sid=None, key=None):
    logmessage("Started observer thread for " + str(sid))
    r = redis.StrictRedis(host=redis_host, port=redis_port, db=redis_offset)
    pubsub = r.pubsub()
    pubsub.subscribe([key, sid])
    for item in pubsub.listen():
        logmessage("2\n" + repr(item))
        if item['type'] != 'message':
            continue
        # logmessage("observer sid: " + str(sid) + ":")
        data = None
        try:
            data = json.loads(item['data'].decode())
        except:
            logmessage("  observer JSON parse error: " + item['data'].decode())
            continue
        if 'message' in data and data['message'] == "KILL" and (('sid' in data and data['sid'] == sid) or 'sid' not in data):
            pubsub.unsubscribe()
            logmessage("  observer unsubscribed and finished for " + str(sid))
            break
        if 'message' in data:
            if data['message'] == "newpage":
                # logmessage("  Got new page for observer")
                try:
                    obj = json.loads(r.get(data['key']).decode())
                except:
                    logmessage("  newpage JSON parse error")
                    continue
                socketio.emit('newpage', {'obj': obj}, namespace='/observer', room=sid)
            elif data['message'] == "start_being_controlled":
                # logmessage("  got start_being_controlled message with key " + str(data['key']))
                socketio.emit('start_being_controlled', {'key': data['key']}, namespace='/observer', room=sid)
        else:
            # logmessage("  Got parameters for observer")
            socketio.emit('pushchanges', {'parameters': data}, namespace='/observer', room=sid)
        logmessage('  exiting observer thread for sid ' + str(sid))


@socketio.on('connect', namespace='/observer')
def on_observer_connect():
    if 'observer' not in session:
        socketio.emit('terminate', {}, namespace='/observer', room=request.sid)
        return
    join_room(request.sid)


@socketio.on('observe', namespace='/observer')
def on_observe(message):
    if 'observer' not in session:
        socketio.emit('terminate', {}, namespace='/observer', room=request.sid)
        return
    if request.sid not in threads:
        key = 'da:input:uid:' + str(message['uid']) + ':i:' + str(message['i']) + ':userid:' + str(message['userid'])
        # logmessage('Observing ' + key)
        threads[request.sid] = socketio.start_background_task(target=observer_thread, sid=request.sid, key=key)


@socketio.on('observerStartControl', namespace='/observer')
def start_control(message):
    if 'observer' not in session:
        socketio.emit('terminate', {}, namespace='/observer', room=request.sid)
        return
    self_key = 'da:control:sid:' + str(request.sid)
    key = 'da:control:uid:' + str(message['uid']) + ':i:' + str(message['i']) + ':userid:' + str(message['userid'])
    existing_sid = rr.get(key)
    if existing_sid is None or existing_sid.decode() == request.sid:
        # logmessage('Controlling ' + key)
        pipe = rr.pipeline()
        pipe.set(self_key, key)
        pipe.expire(self_key, 12)
        pipe.set(key, request.sid)
        pipe.expire(key, 12)
        pipe.execute()
        int_key = 'da:interviewsession:uid:' + str(message['uid']) + ':i:' + str(message['i']) + ':userid:' + str(message['userid'])
        int_sid = rr.get(int_key)
        if int_sid is not None:
            int_sid = int_sid.decode()
            rr.publish(int_sid, json.dumps(dict(messagetype='controllerstart')))
    else:
        logmessage('That key ' + key + ' is already taken')
        key = 'da:session:uid:' + str(message['uid']) + ':i:' + str(message['i']) + ':userid:' + str(message['userid'])
        # rr.publish('da:monitor', json.dumps(dict(messagetype='abortcontroller', key=key)))
        socketio.emit('abortcontrolling', {'key': key}, namespace='/observer', room=request.sid)


@socketio.on('observerStopControl', namespace='/observer')
def stop_control(message):
    if 'observer' not in session:
        socketio.emit('terminate', {}, namespace='/observer', room=request.sid)
        return
    self_key = 'da:control:sid:' + str(request.sid)
    key = 'da:control:uid:' + str(message['uid']) + ':i:' + str(message['i']) + ':userid:' + str(message['userid'])
    logmessage('Stop controlling ' + key)
    existing_sid = rr.get(key)
    pipe = rr.pipeline()
    pipe.delete(self_key)
    if existing_sid is not None and existing_sid.decode() == request.sid:
        pipe.delete(key)
        pipe.execute()
        sid = rr.get('da:interviewsession:uid:' + str(message['uid']) + ':i:' + str(message['i']) + ':userid:' + str(message['userid']))
        if sid is not None:
            sid = sid.decode()
            logmessage("Calling controllerexit 1")
            rr.publish(sid, json.dumps(dict(messagetype='controllerexit', sid=request.sid)))
    else:
        pipe.execute()


@socketio.on('observerChanges', namespace='/observer')
def observer_changes(message):
    logmessage('observerChanges')
    if 'observer' not in session:
        socketio.emit('terminate', {}, namespace='/observer', room=request.sid)
        return
    sid = rr.get('da:interviewsession:uid:' + str(message['uid']) + ':i:' + str(message['i']) + ':userid:' + str(message['userid']))
    if sid is None:
        key = 'da:session:uid:' + str(message['uid']) + ':i:' + str(message['i']) + ':userid:' + str(message['userid'])
        logmessage('observerChanges: sid is none.')
        if rr.get(key) is None:
            logmessage('observerChanges: session has gone away for good.  Sending stopcontrolling.')
            socketio.emit('stopcontrolling', {'key': key}, namespace='/observer', room=request.sid)
        else:
            socketio.emit('noconnection', {'key': key}, namespace='/observer', room=request.sid)
    else:
        sid = sid.decode()
        logmessage('observerChanges: sid exists at ' + time.strftime("%Y-%m-%d %H:%M:%S"))
        rr.publish(sid, json.dumps(dict(messagetype='controllerchanges', sid=request.sid, clicked=message.get('clicked', None), parameters=message['parameters'])))
        # sid=request.sid, yaml_filename=str(message['i']), uid=str(message['uid']), user_id=str(message['userid'])
        self_key = 'da:control:sid:' + str(request.sid)
        key = 'da:control:uid:' + str(message['uid']) + ':i:' + str(message['i']) + ':userid:' + str(message['userid'])
        # logmessage('Controlling ' + key)
        pipe = rr.pipeline()
        pipe.set(self_key, key)
        pipe.expire(key, 12)
        pipe.set(key, request.sid)
        pipe.expire(key, 12)
        pipe.execute()


@socketio.on('disconnect', namespace='/observer')
def on_observer_disconnect():
    logmessage('Client disconnected from observer')
    self_key = 'da:control:sid:' + str(request.sid)
    int_key = rr.get(self_key)
    if int_key is not None:
        int_key = int_key.decode()
        rr.delete(int_key)
        other_sid = rr.get(re.sub(r'^da:control:uid:', 'da:interviewsession:uid:', int_key))
    else:
        other_sid = None
    rr.delete(self_key)
    if other_sid is not None:
        other_sid = other_sid.decode()
        logmessage("Calling controllerexit 2")
        rr.publish(other_sid, json.dumps(dict(messagetype='controllerexit', sid=request.sid)))
    rr.publish(request.sid, json.dumps(dict(message='KILL', sid=request.sid)))


@socketio.on('terminate', namespace='/observer')
def terminate_observer_connection():
    logmessage("terminate_observer_connection")
    # hopefully the disconnect will be triggered
    # if request.sid in threads:
    #     rr.publish(request.sid, json.dumps(dict(origin='client', message='KILL', sid=request.sid)))
    socketio.emit('terminate', {}, namespace='/observer', room=request.sid)
    # disconnect()

if __name__ == '__main__':
    if daconfig.get('expose websockets', False):
        try:
            if 'websockets ip' in daconfig and daconfig['websockets ip']:
                host = daconfig['websockets ip']
            else:
                ifaces = [iface for iface in ni.interfaces() if iface != 'lo']
                host = ni.ifaddresses(ifaces[0])[ni.AF_INET][0]['addr']
            socketio.run(app, host=host, port=daconfig.get('websockets port', 5000))
        except:
            logmessage("Could not find the external IP address")
            if 'websockets ip' in daconfig and daconfig['websockets ip']:
                socketio.run(app, host=daconfig['websockets ip'], port=daconfig.get('websockets port', 5000))
            elif 'websockets port' in daconfig and daconfig['websockets port']:
                socketio.run(app, port=daconfig['websockets port'])
            else:
                socketio.run(app)
    else:
        if 'websockets ip' in daconfig and daconfig['websockets ip']:
            socketio.run(app, host=daconfig['websockets ip'], port=daconfig.get('websockets port', daconfig.get('websockets port', 5000)))
        elif 'websockets port' in daconfig and daconfig['websockets port']:
            socketio.run(app, port=daconfig['websockets port'])
        else:
            socketio.run(app)



__all__ = ['system_packages']

system_packages = [
    "docassemble",
    "docassemble.base",
    "docassemble.demo",
    "3to2",
    "acme",
    "airtable-python-wrapper",
    "alembic",
    "amqp",
    "asn1crypto",
    "astunparse",
    "async-generator",
    "async-timeout",
    "atomicwrites",
    "attrs",
    "azure-common",
    "azure-core",
    "azure-identity",
    "azure-keyvault-secrets",
    "azure-nspkg",
    "azure-storage-blob",
    "Babel",
    "backports.zoneinfo",
    "bcrypt",
    "beautifulsoup4",
    "behave",
    "bidict",
    "billiard",
    "bleach",
    "blinker",
    "boto",
    "boto3",
    "botocore",
    "cachetools",
    "cairocffi",
    "CairoSVG",
    "celery",
    "certbot",
    "certbot-apache",
    "certbot-nginx",
    "certifi",
    "cffi",
    "chardet",
    "charset-normalizer",
    "click",
    "click-didyoumean",
    "click-plugins",
    "click-repl",
    "clicksend-client",
    "colorama",
    "commonmark",
    "ConfigArgParse",
    "configobj",
    "configparser",
    "convertapi",
    "crayons",
    "cryptography",
    "cssselect2",
    "da-pkg-resources",
    "defusedxml",
    "Deprecated",
    "distro",
    "dnspython",
    "docassemble-backports",
    "Docassemble-Flask-User",
    "Docassemble-Pattern",
    "docassemble-textstat",
    "docassemblekvsession",
    "docopt",
    "docutils",
    "docxcompose",
    "docxtpl",
    "email-validator",
    "et-xmlfile",
    "eventlet",
    "Flask",
    "Flask-Babel",
    "Flask-Cors",
    "Flask-Login",
    "Flask-Mail",
    "Flask-SocketIO",
    "Flask-SQLAlchemy",
    "Flask-WTF",
    "future",
    "gcs-oauth2-boto-plugin",
    "geographiclib",
    "geopy",
    "google-api-core",
    "google-api-python-client",
    "google-auth",
    "google-auth-httplib2",
    "google-auth-oauthlib",
    "google-cloud-core",
    "google-cloud-storage",
    "google-cloud-translate",
    "google-cloud-vision",
    "google-crc32c",
    "google-i18n-address",
    "google-reauth",
    "google-resumable-media",
    "googleapis-common-protos",
    "greenlet",
    "grpcio",
    "grpcio-status",
    "gspread",
    "guess-language-spirit",
    "h11",
    "httplib2",
    "humanize",
    "Hyphenate",
    "idna",
    "importlib-metadata",
    "importlib-resources",
    "iniconfig",
    "iso8601",
    "isodate",
    "itsdangerous",
    "jdcal",
    "jeepney",
    "jellyfish",
    "Jinja2",
    "jmespath",
    "joblib",
    "josepy",
    "keyring",
    "kombu",
    "libcst",
    "links-from-link-header",
    "lxml",
    "Mako",
    "Markdown",
    "MarkupSafe",
    "mdx-smartypants",
    "minio",
    "mod_wsgi",
    "mod-wsgi",
    "monotonic",
    "msal",
    "msal-extensions",
    "msrest",
    "mypy-extensions",
    "namedentities",
    "netifaces",
    "nltk",
    "num2words",
    "numpy",
    "oauth2client",
    "oauthlib",
    "openpyxl",
    "ordered-set",
    "outcome",
    "packaging",
    "pandas",
    "parse",
    "parse-type",
    "parsedatetime",
    "passlib",
    "pathlib",
    "pdfminer.six",
    "phonenumbers",
    "Pillow",
    "pip",
    "pkginfo",
    "pluggy",
    "ply",
    "portalocker",
    "prompt-toolkit",
    "proto-plus",
    "protobuf",
    "psutil",
    "psycopg2-binary",
    "py",
    "pyasn1",
    "pyasn1-modules",
    "pycountry",
    "pycparser",
    "pycryptodome",
    "pycryptodomex",
    "pycurl",
    "Pygments",
    "PyJWT",
    "PyLaTeX",
    "PyNaCl",
    "PyOpenSSL",
    "pyOpenSSL",
    "pyotp",
    "pyparsing",
    "pyPdf",
    "PyPDF2",
    "pypdftk",
    "pypng",
    "pyRFC3339",
    "PySocks",
    "pytest",
    "python-augeas",
    "python-dateutil",
    "python-docx",
    "python-editor",
    "python-engineio",
    "python-http-client",
    "python-ldap",
    "python-socketio",
    "pytz",
    "pytz-deprecation-shim",
    "pyu2f",
    "PyYAML",
    "pyzbar",
    "qrcode",
    "rauth",
    "readme-renderer",
    "redis",
    "regex",
    "reportlab",
    "repoze.lru",
    "requests",
    "requests-oauthlib",
    "requests-toolbelt",
    "retry-decorator",
    "rfc3339",
    "rfc3986",
    "rich",
    "rsa",
    "ruamel.yaml",
    "ruamel.yaml.clib",
    "s3transfer",
    "s4cmd",
    "scikit-learn",
    "scipy",
    "SecretStorage",
    "selenium",
    "sendgrid",
    "setuptools",
    "simplekv",
    "six",
    "sklearn",
    "sniffio",
    "SocksiPy-branch",
    "sortedcontainers",
    "soupsieve",
    "SQLAlchemy",
    "starkbank-ecdsa",
    "tailer",
    "telnyx",
    "threadpoolctl",
    "tinycss2",
    "titlecase",
    "toml",
    "tomli",
    "tqdm",
    "trio",
    "trio-websocket",
    "twilio",
    "twine",
    "typing-extensions",
    "typing-inspect",
    "tzdata",
    "tzlocal",
    "ua-parser",
    "uritemplate",
    "urllib3",
    "us",
    "user-agents",
    "uWSGI",
    "vine",
    "wcwidth",
    "webdriver-manager",
    "webencodings",
    "Werkzeug",
    "wheel",
    "wrapt",
    "wsproto",
    "WTForms",
    "xfdfgen",
    "xlrd",
    "XlsxWriter",
    "xlwt",
    "zipp",
    "zope.component",
    "zope.event",
    "zope.hookable",
    "zope.interface"
]



import os
import sys
import socket
import tempfile
import subprocess
import xmlrpc.client
import re
# from io import StringIO
import shutil
import time
import fcntl
from packaging import version
import docassemble.base.config
from docassemble.base.config import daconfig
from docassemble.base.logger import logmessage
from docassemble.webapp.info import system_packages

installed_distribution_cache = None

if __name__ == "__main__":
    docassemble.base.config.load(arguments=sys.argv)
    if 'initialize' in sys.argv:
        mode = 'initialize'
    elif 'check_for_updates' in sys.argv:
        mode = 'check_for_updates'
    else:
        mode = 'initialize'

USING_SUPERVISOR = bool(os.environ.get('SUPERVISOR_SERVER_URL', None))
SINGLE_SERVER = USING_SUPERVISOR and bool(':all:' in ':' + os.environ.get('CONTAINERROLE', 'all') + ':')


def fix_fnctl():
    try:
        flags = fcntl.fcntl(sys.stdout, fcntl.F_GETFL)
        fcntl.fcntl(sys.stdout, fcntl.F_SETFL, flags & ~os.O_NONBLOCK)
        logmessage("fix_fnctl: updated stdout")
    except:
        pass
    try:
        flags = fcntl.fcntl(sys.stderr, fcntl.F_GETFL)
        fcntl.fcntl(sys.stderr, fcntl.F_SETFL, flags & ~os.O_NONBLOCK)
        logmessage("fix_fnctl: updated stderr")
    except:
        pass


def remove_inactive_hosts(start_time=None):
    if start_time is None:
        start_time = time.time()
    logmessage("remove_inactive_hosts: starting after " + str(time.time() - start_time) + " seconds")
    if USING_SUPERVISOR:
        from docassemble.base.config import hostname  # pylint: disable=import-outside-toplevel
        # from docassemble.webapp.app_object import app  # pylint: disable=import-outside-toplevel
        from docassemble.webapp.db_object import db  # pylint: disable=import-outside-toplevel
        from docassemble.webapp.core.models import Supervisors  # pylint: disable=import-outside-toplevel
        from sqlalchemy import select, delete  # pylint: disable=import-outside-toplevel
        to_delete = set()
        for host in db.session.execute(select(Supervisors)).scalars():
            if host.hostname == hostname:
                continue
            try:
                socket.gethostbyname(host.hostname)
                server = xmlrpc.client.Server(host.url + '/RPC2')
                server.supervisor.getState()
            except:
                to_delete.add(host.id)
        for id_to_delete in to_delete:
            db.session.execute(delete(Supervisors).filter_by(id=id_to_delete))
            db.session.commit()
    logmessage("remove_inactive_hosts: ended after " + str(time.time() - start_time) + " seconds")


class DummyPackage:

    def __init__(self, name):
        self.name = name
        self.type = 'pip'
        self.limitation = None


def clear_invalid_package_auth(start_time=None):
    if start_time is None:
        start_time = time.time()
    logmessage("clear_invalid_package_auth: starting after " + str(time.time() - start_time) + " seconds")
    from docassemble.webapp.db_object import db  # pylint: disable=import-outside-toplevel
    from docassemble.webapp.packages.models import PackageAuth  # pylint: disable=import-outside-toplevel
    from sqlalchemy import delete  # pylint: disable=import-outside-toplevel
    db.session.execute(delete(PackageAuth).filter_by(package_id=None))
    db.session.commit()
    logmessage("clear_invalid_package_auth: finishing after " + str(time.time() - start_time) + " seconds")


def check_for_updates(start_time=None, invalidate_cache=True, full=True):
    if start_time is None:
        start_time = time.time()
    logmessage("check_for_updates: starting after " + str(time.time() - start_time) + " seconds")
    if invalidate_cache:
        invalidate_installed_distributions_cache()
    from docassemble.base.config import hostname  # pylint: disable=import-outside-toplevel
    # from docassemble.webapp.app_object import app  # pylint: disable=import-outside-toplevel
    from docassemble.webapp.db_object import db  # pylint: disable=import-outside-toplevel
    from docassemble.webapp.packages.models import Package, Install, PackageAuth  # pylint: disable=import-outside-toplevel
    from sqlalchemy import select, delete  # pylint: disable=import-outside-toplevel
    ok = True  # tracks whether there have been any errors
    here_already = {}  # packages that are installed on this server, based on pip list. package name -> package version
    results = {}  # result of actions taken on a package. package name -> message
    if full:
        logmessage("check_for_updates: 0.5 after " + str(time.time() - start_time) + " seconds")
        for package_name in ('psycopg2', 'pdfminer', 'pdfminer3k', 'py-bcrypt', 'pycrypto', 'constraint', 'distutils2', 'azure-storage', 'Flask-User', 'Marisol'):
            result = db.session.execute(delete(Package).filter_by(name=package_name))
            if result.rowcount > 0:
                db.session.commit()
        logmessage("check_for_updates: 1 after " + str(time.time() - start_time) + " seconds")
    installed_packages = get_installed_distributions(start_time=start_time)
    for package in installed_packages:
        here_already[package.key] = package.version
    changed = False
    # clean up old packages that cause problems.
    if full:
        if 'psycopg2' in here_already:
            logmessage("check_for_updates: uninstalling psycopg2")
            uninstall_package(DummyPackage('psycopg2'), start_time=start_time)
            if 'psycopg2-binary' in here_already:
                logmessage("check_for_updates: reinstalling psycopg2-binary")
                uninstall_package(DummyPackage('psycopg2-binary'), start_time=start_time)
                install_package(DummyPackage('psycopg2-binary'), start_time=start_time)
            changed = True
        if 'psycopg2-binary' not in here_already:
            logmessage("check_for_updates: installing psycopg2-binary")
            install_package(DummyPackage('psycopg2-binary'), start_time=start_time)
            changed = True
        if 'kombu' not in here_already or version.parse(here_already['kombu']) <= version.parse('4.1.0'):
            logmessage("check_for_updates: installing new kombu version")
            install_package(DummyPackage('kombu'), start_time=start_time)
            changed = True
        if 'celery' not in here_already or version.parse(here_already['celery']) <= version.parse('4.1.0'):
            logmessage("check_for_updates: installing new celery version")
            install_package(DummyPackage('celery'), start_time=start_time)
            changed = True
        if 'pycrypto' in here_already:
            logmessage("check_for_updates: uninstalling pycrypto")
            uninstall_package(DummyPackage('pycrypto'), start_time=start_time)
            if 'pycryptodome' in here_already:
                logmessage("check_for_updates: reinstalling pycryptodome")
                uninstall_package(DummyPackage('pycryptodome'), start_time=start_time)
                install_package(DummyPackage('pycryptodome'), start_time=start_time)
            changed = True
        if 'pycryptodome' not in here_already:
            logmessage("check_for_updates: installing pycryptodome")
            install_package(DummyPackage('pycryptodome'), start_time=start_time)
            changed = True
        if 'pdfminer' in here_already:
            logmessage("check_for_updates: uninstalling pdfminer")
            uninstall_package(DummyPackage('pdfminer'), start_time=start_time)
            changed = True
        if 'Marisol' in here_already:
            logmessage("check_for_updates: uninstalling Marisol")
            uninstall_package(DummyPackage('Marisol'), start_time=start_time)
            changed = True
        if 'pdfminer3k' in here_already:
            logmessage("check_for_updates: uninstalling pdfminer3k")
            uninstall_package(DummyPackage('pdfminer3k'), start_time=start_time)
            changed = True
        if 'pdfminer.six' in here_already:
            try:
                from pdfminer.pdfparser import PDFParser  # noqa: F401 # pylint: disable=import-outside-toplevel,unused-import
                from pdfminer.pdfdocument import PDFDocument  # noqa: F401 # pylint: disable=import-outside-toplevel,unused-import
            except:
                logmessage("check_for_updates: reinstalling pdfminer.six")
                uninstall_package(DummyPackage('pdfminer.six'), start_time=start_time)
                install_package(DummyPackage('pdfminer.six'), start_time=start_time)
        else:
            logmessage("check_for_updates: installing pdfminer.six")
            install_package(DummyPackage('pdfminer.six'), start_time=start_time)
            changed = True
        if 'py-bcrypt' in here_already:
            logmessage("check_for_updates: uninstalling py-bcrypt")
            uninstall_package(DummyPackage('py-bcrypt'), start_time=start_time)
            changed = True
            if 'bcrypt' in here_already:
                logmessage("check_for_updates: reinstalling bcrypt")
                uninstall_package(DummyPackage('bcrypt'), start_time=start_time)
                install_package(DummyPackage('bcrypt'), start_time=start_time)
                changed = True
        if 'bcrypt' not in here_already:
            logmessage("check_for_updates: installing bcrypt")
            install_package(DummyPackage('bcrypt'), start_time=start_time)
            changed = True
        if changed:
            installed_packages = get_installed_distributions(start_time=start_time)
            here_already = {}
            for package in installed_packages:
                here_already[package.key] = package.version
    logmessages = ''
    packages = {}  # packages that the database says are active and that have a type; package id -> package row
    installs = {}  # install rows representing what the database says in installed; package id -> install row
    to_install = []  # package rows of packages to install
    to_uninstall = []  # package rows of packages to uninstall
    system_packages_to_fix = []
    uninstall_done = set()  # set of package names of packages already uninstalled
    uninstalled_packages = {}  # packages with active=False; package id -> package row
    package_ids_to_delete = set()  # set if IDs of package rows that have active=False but there is another row with active= True
    package_by_name = {}  # packages the database says are active; package name -> package row
    logmessage("check_for_updates: 2 after " + str(time.time() - start_time) + " seconds")
    # create dict package_by_name that maps a package name to a package database row,
    # for packages that the database says should be installed
    for package in db.session.execute(select(Package.name).filter_by(active=True)):
        package_by_name[package.name] = package
        # logmessage("check_for_updates: database includes a package called " + package.name + " after " + str(time.time() - start_time) + " seconds")
    # packages is what is supposed to be installed
    logmessage("check_for_updates: 3 after " + str(time.time() - start_time) + " seconds")
    # create dict packages that maps a package ID to a package database row if the type is not null
    for package in db.session.execute(select(Package).filter_by(active=True)).scalars():
        if package.type is not None:
            packages[package.id] = package
            # logmessage("check_for_updates: database includes a package called " + package.name + " that has a type after " + str(time.time() - start_time) + " seconds")
            # print("Found a package " + package.name)
    logmessage("check_for_updates: 4 after " + str(time.time() - start_time) + " seconds")
    # create dict uninstalled_packages that maps a package ID to a package database row if SQL wants the package deleted
    for package in db.session.execute(select(Package).filter_by(active=False)).scalars():
        # don't uninstall a system package
        if package.name in system_packages:
            logmessages += "Not uninstalling " + str(package.name) + " because it is a system package."
            system_packages_to_fix.append(package)
            continue
        if package.name in package_by_name:
            # there are two Package entries for the same package, one with active=True, one with active=False
            logmessage("check_for_updates: conflicting package entries for " + package.name + " after " + str(time.time() - start_time) + " seconds")
            # let's delete the one with active=False
            package_ids_to_delete.add(package.id)
        else:
            # logmessage("check_for_updates: database says " + package.name + " should be uninstalled after " + str(time.time() - start_time) + " seconds")
            uninstalled_packages[package.id] = package  # this is what the database says should be uninstalled
    logmessage("check_for_updates: 5 after " + str(time.time() - start_time) + " seconds")
    # build to_uninstall, which is the list of Package database entries targeted for uninstallation.
    # Only add if there is an Install entry linked to the package id and there is no active Package row
    # with the same name.
    # Also build installs, which is a dict mapping package IDs to install rows, based on
    # what the Install table says
    for install in db.session.execute(select(Install).filter_by(hostname=hostname)).scalars():
        installs[install.package_id] = install  # this is what the database says in installed on this server
        if install.package_id in uninstalled_packages and uninstalled_packages[install.package_id].name not in package_by_name:
            logmessage("check_for_updates: " + uninstalled_packages[install.package_id].name + " will be uninstalled after " + str(time.time() - start_time) + " seconds")
            to_uninstall.append(uninstalled_packages[install.package_id])  # uninstall if it is installed
    changed = False
    package_owner = {}
    logmessage("check_for_updates: 6 after " + str(time.time() - start_time) + " seconds")
    for auth in db.session.execute(select(PackageAuth).filter_by(authtype='owner')).scalars():
        package_owner[auth.package_id] = auth.user_id
    logmessage("check_for_updates: 7 after " + str(time.time() - start_time) + " seconds")
    # Add Install rows for any active Package rows with non-null types if the package is present on the server
    # Also add to the installs dict after adding.
    for package in packages.values():
        if package.id not in installs and package.name in here_already:
            logmessage("check_for_updates: package " + package.name + " here already.  Writing an Install record for it.")
            install = Install(hostname=hostname, packageversion=here_already[package.name], version=package.version, package_id=package.id)
            db.session.add(install)
            installs[package.id] = install
            changed = True
    if changed:
        db.session.commit()
    logmessage("check_for_updates: 8 after " + str(time.time() - start_time) + " seconds")
    for package in packages.values():
        # logmessage("check_for_updates: processing package id " + str(package.id))
        # logmessage("1: " + str(installs[package.id].packageversion) + " 2: " + str(package.packageversion))
        if (package.packageversion is not None and package.id in installs and installs[package.id].packageversion is None) or (package.packageversion is not None and package.id in installs and installs[package.id].packageversion is not None and version.parse(package.packageversion) > version.parse(installs[package.id].packageversion)):
            logmessage("check_for_updates: a new version of " + package.name + " is needed because the necessary package version, " + str(package.packageversion) + ", is ahead of the installed version, " + str(installs[package.id].packageversion) + " after " + str(time.time() - start_time) + " seconds")
            new_version_needed = True
        else:
            new_version_needed = False
        # logmessage("got here and new version is " + str(new_version_needed))
        # Check for missing local packages
        if (package.name not in here_already) and (package.id in installs):
            logmessage("check_for_updates: the package " + package.name + " is supposed to be installed on this server, but was not detected after " + str(time.time() - start_time) + " seconds")
            package_missing = True
        else:
            package_missing = False
        if package.id in installs and package.version > installs[package.id].version:
            logmessage("check_for_updates: the package " + package.name + " has internal version " + str(package.version) + " but the installed version has version " + str(installs[package.id].version) + " after " + str(time.time() - start_time) + " seconds")
            package_version_greater = True
        else:
            package_version_greater = False
        if package.id not in installs:
            logmessage("check_for_updates: the package " + package.name + " is not in the table of installed packages for this server after " + str(time.time() - start_time) + " seconds")
        if package.id not in installs or package_version_greater or new_version_needed or package_missing:
            if package.name in system_packages:
                logmessages += "Not upgrading " + str(package.name) + " because it is a system package and its version needs to be consistent with the version of the package that is required by the docassemble.webapp package."
                logmessage("check_for_updates: the package " + package.name + " is a system package and cannot be updated except through docassemble.webapp after " + str(time.time() - start_time) + " seconds")
                system_packages_to_fix.append(package)
            else:
                to_install.append(package)
    # logmessage("done with that")
    logmessage("check_for_updates: 9 after " + str(time.time() - start_time) + " seconds")
    for package in to_uninstall:
        # logmessage("Going to uninstall a package: " + package.name)
        if package.name in uninstall_done:
            logmessage("check_for_updates: skipping uninstallation of " + str(package.name) + " because already uninstalled after " + str(time.time() - start_time) + " seconds")
            continue
        if package.name not in here_already:
            logmessage("check_for_updates: skipping uninstallation of " + str(package.name) + " because not installed" + " after " + str(time.time() - start_time) + " seconds")
            returnval = 1
            newlog = ''
        else:
            logmessage("check_for_updates: calling uninstall_package on " + package.name)
            returnval, newlog = uninstall_package(package, start_time=start_time)
        uninstall_done.add(package.name)
        logmessages += newlog
        if returnval == 0:
            db.session.execute(delete(Install).filter_by(hostname=hostname, package_id=package.id))
            results[package.name] = 'pip uninstall command returned success code.  See log for details.'
        elif returnval == 1:
            db.session.execute(delete(Install).filter_by(hostname=hostname, package_id=package.id))
            results[package.name] = 'pip uninstall was not run because the package was not installed.'
        else:
            results[package.name] = 'pip uninstall command returned failure code'
            ok = False
    packages_to_delete = []
    logmessage("check_for_updates: 10 after " + str(time.time() - start_time) + " seconds")
    did_something = False
    for package in to_install:
        did_something = True
        logmessage("check_for_updates: going to install a package: " + package.name + " after " + str(time.time() - start_time) + " seconds")
        # if doing_startup and package.name.startswith('docassemble') and package.name in here_already:
        #     # adding this because of unpredictability of installing new versions of docassemble
        #     # just because of a system restart.
        #     logmessage("check_for_updates: skipping update on " + str(package.name))
        #     continue
        returnval, newlog = install_package(package, start_time=start_time)
        logmessages += newlog
        logmessage("check_for_updates: return value was " + str(returnval) + " after " + str(time.time() - start_time) + " seconds")
        if returnval != 0:
            logmessage("Return value was not good" + " after " + str(time.time() - start_time) + " seconds")
            ok = False
        # pip._vendor.pkg_resources._initialize_master_working_set()
        if full:
            pip_info = get_pip_info(package.name, start_time=start_time)
            real_name = pip_info['Name']
            logmessage("check_for_updates: real name of package " + str(package.name) + " is " + str(real_name) + " after " + str(time.time() - start_time) + " seconds")
        else:
            real_name = package.name
        if real_name is None:
            results[package.name] = 'install failed'
            ok = False
            if package.name not in here_already:
                logmessage("check_for_updates: removing package entry for " + package.name + " after " + str(time.time() - start_time) + " seconds")
                packages_to_delete.append(package)
        elif returnval != 0:
            results[package.name] = 'pip install command returned failure code'
        else:
            results[package.name] = 'pip install command returned success code.  See log for details.'
            if real_name != package.name:
                logmessage("check_for_updates: changing name" + " after " + str(time.time() - start_time) + " seconds")
                package.name = real_name
            if package.id in installs:
                install = installs[package.id]
                install.version = package.version
            else:
                install = Install(hostname=hostname, packageversion=package.packageversion, version=package.version, package_id=package.id)
                db.session.add(install)
            db.session.commit()
    if did_something:
        update_versions(start_time=start_time)
        if full and add_dependencies(package_owner.get(package.id, 1), start_time=start_time):
            update_versions(start_time=start_time)
    logmessage("check_for_updates: 11 after " + str(time.time() - start_time) + " seconds")
    for package in packages_to_delete:
        db.session.delete(package)
    logmessage("check_for_updates: 12 after " + str(time.time() - start_time) + " seconds")
    db.session.commit()
    logmessage("check_for_updates: finished uninstalling and installing after " + str(time.time() - start_time) + " seconds")
    return ok, logmessages, results


def update_versions(start_time=None):
    if start_time is None:
        start_time = time.time()
    logmessage("update_versions: starting after " + str(time.time() - start_time) + " seconds")
    from docassemble.base.config import hostname  # pylint: disable=import-outside-toplevel
    # from docassemble.webapp.app_object import app  # pylint: disable=import-outside-toplevel
    from docassemble.webapp.db_object import db  # pylint: disable=import-outside-toplevel
    from docassemble.webapp.packages.models import Package, Install  # pylint: disable=import-outside-toplevel
    from sqlalchemy import select  # pylint: disable=import-outside-toplevel
    install_by_id = {}
    for install in db.session.execute(select(Install).filter_by(hostname=hostname)).scalars():
        install_by_id[install.package_id] = install
    package_by_name = {}
    for package in db.session.execute(select(Package).filter_by(active=True).order_by(Package.name, Package.id.desc())).scalars():
        if package.name in package_by_name:
            continue
        package_by_name[package.name] = Object(id=package.id, packageversion=package.packageversion, name=package.name)
    installed_packages = get_installed_distributions(start_time=start_time)
    for package in installed_packages:
        if package.key in package_by_name:
            if package_by_name[package.key].id in install_by_id and package.version != install_by_id[package_by_name[package.key].id].packageversion:
                for install_row in db.session.execute(select(Install).filter_by(hostname=hostname, package_id=package_by_name[package.key].id)).scalars():
                    install_row.packageversion = package.version
            if package.version != package_by_name[package.key].packageversion:
                for package_row in db.session.execute(select(Package).filter_by(active=True, name=package_by_name[package.key].name).with_for_update()).scalars():
                    package_row.packageversion = package.version
    db.session.commit()
    logmessage("update_versions: ended after " + str(time.time() - start_time))


def get_home_page_dict():
    PACKAGE_DIRECTORY = daconfig.get('packages', '/usr/share/docassemble/local' + str(sys.version_info.major) + '.' + str(sys.version_info.minor))
    FULL_PACKAGE_DIRECTORY = os.path.join(PACKAGE_DIRECTORY, 'lib', 'python' + str(sys.version_info.major) + '.' + str(sys.version_info.minor), 'site-packages')
    home_page = {}
    for d in os.listdir(FULL_PACKAGE_DIRECTORY):
        if not d.startswith('docassemble.'):
            continue
        metadata_path = os.path.join(d, 'METADATA')
        if os.path.isfile(metadata_path):
            name = None
            url = None
            with open(metadata_path, 'r', encoding='utf-8') as fp:
                for line in fp:
                    if line.startswith('Name: '):
                        name = line[6:]
                    elif line.startswith('Home-page: '):
                        url = line[11:].rstrip('/')
                        break
            if name:
                home_page[name.lower()] = url
    return home_page


def add_dependencies(user_id, start_time=None):
    if start_time is None:
        start_time = time.time()
    # logmessage('add_dependencies: user_id is ' + str(user_id))
    logmessage("add_dependencies: starting after " + str(time.time() - start_time) + " seconds")
    from docassemble.base.config import hostname  # pylint: disable=import-outside-toplevel
    # from docassemble.webapp.app_object import app  # pylint: disable=import-outside-toplevel
    from docassemble.webapp.db_object import db  # pylint: disable=import-outside-toplevel
    from docassemble.webapp.packages.models import Package, Install, PackageAuth  # pylint: disable=import-outside-toplevel
    from sqlalchemy import select, delete  # pylint: disable=import-outside-toplevel
    packages_known = set()
    for package in db.session.execute(select(Package.name).filter_by(active=True)):
        packages_known.add(package.name)
    installed_packages = get_installed_distributions(start_time=start_time)
    home_pages = None
    packages_to_add = []
    for package in installed_packages:
        if package.key in packages_known:
            continue
        if package.key.startswith('mysqlclient') or package.key.startswith('mysql-connector') or package.key.startswith('MySQL-python'):
            continue
        db.session.execute(delete(Package).filter_by(name=package.key))
        packages_to_add.append(package)
    did_something = False
    if len(packages_to_add) > 0:
        did_something = True
        db.session.commit()
        for package in packages_to_add:
            if package.key.startswith('docassemble.'):
                if home_pages is None:
                    home_pages = get_home_page_dict()
                home_page = home_pages.get(package.key.lower(), None)
                if home_page is not None and re.search(r'/github.com/', home_page):
                    package_entry = Package(name=package.key, package_auth=PackageAuth(user_id=user_id), type='git', giturl=home_page, packageversion=package.version, dependency=True)
                else:
                    package_entry = Package(name=package.key, package_auth=PackageAuth(user_id=user_id), type='pip', packageversion=package.version, dependency=True)
            else:
                package_entry = Package(name=package.key, package_auth=PackageAuth(user_id=user_id), type='pip', packageversion=package.version, dependency=True)
            db.session.add(package_entry)
            db.session.commit()
            install = Install(hostname=hostname, packageversion=package_entry.packageversion, version=package_entry.version, package_id=package_entry.id)
            db.session.add(install)
            db.session.commit()
    logmessage("add_dependencies: ending after " + str(time.time() - start_time) + " seconds")
    return did_something


def fix_names():
    # from docassemble.webapp.app_object import app  # pylint: disable=import-outside-toplevel
    from docassemble.webapp.db_object import db  # pylint: disable=import-outside-toplevel
    from docassemble.webapp.packages.models import Package  # pylint: disable=import-outside-toplevel
    from sqlalchemy import select  # pylint: disable=import-outside-toplevel
    installed_packages = [package.key for package in get_installed_distributions()]
    for package in db.session.execute(select(Package).filter_by(active=True).with_for_update()).scalars():
        if package.name not in installed_packages:
            pip_info = get_pip_info(package.name)
            actual_name = pip_info['Name']
            if actual_name is not None:
                package.name = actual_name
            else:
                logmessage("fix_names: package " + package.name + " does not appear to be installed")
    db.session.commit()


def splitall(path):
    allparts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:
            allparts.insert(0, parts[0])
            break
        if parts[1] == path:
            allparts.insert(0, parts[1])
            break
        path = parts[0]
        allparts.insert(0, parts[1])
    return allparts


def install_package(package, start_time=None):
    if start_time is None:
        start_time = time.time()
    invalidate_installed_distributions_cache()
    logmessage("install_package: " + package.name + " after " + str(time.time() - start_time) + " seconds")
    if package.type == 'zip' and package.upload is None:
        return 0, ''
    from docassemble.webapp.files import SavedFile  # pylint: disable=import-outside-toplevel
    PACKAGE_DIRECTORY = daconfig.get('packages', '/usr/share/docassemble/local' + str(sys.version_info.major) + '.' + str(sys.version_info.minor))
    logfilecontents = ''
    pip_log = tempfile.NamedTemporaryFile()
    temp_dir = tempfile.mkdtemp(prefix='SavedFile')
    # use_pip_cache = r.get('da:updatepackage:use_pip_cache')
    # if use_pip_cache is None:
    #     disable_pip_cache = False
    # elif int(use_pip_cache):
    #     disable_pip_cache = False
    # else:
    #     disable_pip_cache = True
    disable_pip_cache = True
    if package.type in ('zip', 'git'):
        logmessage("install_package: calling uninstall_package on " + package.name + " after " + str(time.time() - start_time) + " seconds")
        returnval, newlog = uninstall_package(package, start_time=start_time)
        logfilecontents += newlog
    if package.type == 'zip' and package.upload is not None:
        saved_file = SavedFile(package.upload, extension='zip', fix=True)
        commands = ['pip', 'install']
        if disable_pip_cache:
            commands.append('--no-cache-dir')
        commands.extend(['--quiet', '--prefix=' + PACKAGE_DIRECTORY, '--src=' + temp_dir, '--log-file=' + pip_log.name, '--upgrade', saved_file.path + '.zip'])
    elif package.type == 'git' and package.giturl:
        if package.gitbranch is not None:
            branchpart = '@' + str(package.gitbranch)
        else:
            branchpart = ''
        if str(package.giturl).endswith('.git'):
            gitsuffix = ''
        else:
            gitsuffix = '.git'
        if str(package.giturl).startswith('git+'):
            gitprefix = ''
        else:
            gitprefix = 'git+'
        if package.gitsubdir is not None:
            commands = ['pip', 'install']
            if disable_pip_cache:
                commands.append('--no-cache-dir')
            commands.extend(['--quiet', '--prefix=' + PACKAGE_DIRECTORY, '--src=' + temp_dir, '--upgrade', '--log-file=' + pip_log.name, gitprefix + str(package.giturl).rstrip('/') + gitsuffix + branchpart + '#egg=' + package.name + '&subdirectory=' + str(package.gitsubdir)])
        else:
            commands = ['pip', 'install']
            if disable_pip_cache:
                commands.append('--no-cache-dir')
            commands.extend(['--quiet', '--prefix=' + PACKAGE_DIRECTORY, '--src=' + temp_dir, '--upgrade', '--log-file=' + pip_log.name, gitprefix + str(package.giturl).rstrip('/') + gitsuffix + branchpart + '#egg=' + package.name])
    elif package.type == 'pip':
        if package.limitation is None:
            limit = ""
        else:
            limit = str(package.limitation)
        commands = ['pip', 'install']
        if disable_pip_cache:
            commands.append('--no-cache-dir')
        commands.extend(['--quiet', '--prefix=' + PACKAGE_DIRECTORY, '--src=' + temp_dir, '--upgrade', '--log-file=' + pip_log.name, package.name + limit])
    else:
        logmessage("install_package: wrong package type after " + str(time.time() - start_time) + " seconds")
        return 1, 'Unable to recognize package type: ' + package.name
    logmessage("install_package: running " + " ".join(commands) + " after " + str(time.time() - start_time) + " seconds")
    logfilecontents += "install_package: running " + " ".join(commands) + "\n"
    returnval = 1
    try:
        subprocess.run(commands, check=False)
        returnval = 0
    except subprocess.CalledProcessError as err:
        returnval = err.returncode
    pip_log.seek(0)
    with open(pip_log.name, 'r', encoding='utf-8') as x:
        logfilecontents += x.read()
    pip_log.close()
    shutil.rmtree(temp_dir)
    logmessage('install_package: returnval is: ' + str(returnval))
    logmessage('install_package: done' + " after " + str(time.time() - start_time) + " seconds")
    return returnval, logfilecontents


def uninstall_package(package, start_time=None):
    if start_time is None:
        start_time = time.time()
    invalidate_installed_distributions_cache()
    logmessage('uninstall_package: ' + package.name + " after " + str(time.time() - start_time) + " seconds")
    logfilecontents = 'uninstall_package: ' + package.name + "\n"
    pip_log = tempfile.NamedTemporaryFile()
    commands = ['pip', 'uninstall', '--yes', '--log-file=' + pip_log.name, package.name]
    logmessage("uninstall_package: running " + " ".join(commands) + " after " + str(time.time() - start_time) + " seconds")
    logfilecontents += "Running " + (" ".join(commands)) + "\n"
    try:
        subprocess.run(commands, check=False)
        returnval = 0
    except subprocess.CalledProcessError as err:
        returnval = err.returncode
    pip_log.seek(0)
    with open(pip_log.name, 'r', encoding='utf-8') as x:
        logfilecontents += x.read()
    pip_log.close()
    logmessage('uninstall_package: returnval is: ' + str(returnval))
    logmessage("uninstall_package: done after " + str(time.time() - start_time) + " seconds")
    logfilecontents += 'returnval is: ' + str(returnval) + "\n"
    logfilecontents += 'uninstall_package: done' + "\n"
    return returnval, logfilecontents


class Object:

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)


def invalidate_installed_distributions_cache():
    global installed_distribution_cache
    installed_distribution_cache = None


def get_installed_distributions(start_time=None):
    global installed_distribution_cache
    if installed_distribution_cache is not None:
        return installed_distribution_cache
    if start_time is None:
        start_time = time.time()
    logmessage("get_installed_distributions: starting after " + str(time.time() - start_time) + " seconds")
    results = []
    # try:
    #     output = subprocess.check_output(['pip', '--version']).decode('utf-8', 'ignore')
    # except subprocess.CalledProcessError as err:
    #     output = err.output.decode('utf-8', 'ignore')
    # logmessage("get_installed_distributions: pip version: " + output.strip() + " after " + str(time.time() - start_time) + " seconds")
    try:
        output = subprocess.check_output(['pip', 'list', '--format=freeze']).decode('utf-8', 'ignore')
    except subprocess.CalledProcessError as err:
        output = err.output.decode('utf-8', 'ignore')
    # logmessage("get_installed_distributions: result of pip list --format freeze was:\n" + str(output))
    for line in output.split('\n'):
        a = line.split("==")
        if len(a) == 2:
            results.append(Object(key=a[0], version=a[1]))
    installed_distribution_cache = results
    logmessage("get_installed_distributions: ending after " + str(time.time() - start_time) + " seconds")
    # logmessage(repr([x.key for x in results]))
    return results


def get_pip_info(package_name, start_time=None):
    if start_time is None:
        start_time = time.time()
    logmessage("get_pip_info: " + package_name + " after " + str(time.time() - start_time) + " seconds")
    try:
        output = subprocess.check_output(['pip', 'show', package_name]).decode('utf-8', 'ignore')
    except subprocess.CalledProcessError as err:
        output = ""
        logmessage("get_pip_info: error.  output was " + err.output.decode('utf-8', 'ignore') + " after " + str(time.time() - start_time) + " seconds")
    # old_stdout = sys.stdout
    # sys.stdout = saved_stdout = StringIO()
    # pip.main(['show', package_name])
    # sys.stdout = old_stdout
    # output = saved_stdout.getvalue()
    results = {}
    if not isinstance(output, str):
        output = output.decode('utf-8', 'ignore')
    for line in output.split('\n'):
        # logmessage("Found line " + str(line))
        a = line.split(": ")
        if len(a) == 2:
            # logmessage("Found " + a[0] + " which was " + a[1])
            results[a[0]] = a[1]
    for key in ['Name', 'Home-page', 'Version']:
        if key not in results:
            results[key] = None
    logmessage("get_pip_info: returning after " + str(time.time() - start_time) + " seconds")
    return results


def main():
    # import docassemble.webapp.database  # pylint: disable=import-outside-toplevel
    start_time = time.time()
    from docassemble.webapp.app_object import app  # pylint: disable=import-outside-toplevel
    with app.app_context():
        from docassemble.webapp.db_object import db  # pylint: disable=import-outside-toplevel
        from docassemble.webapp.packages.models import Package  # pylint: disable=import-outside-toplevel
        from sqlalchemy import select  # pylint: disable=import-outside-toplevel
        # app.config['SQLALCHEMY_DATABASE_URI'] = docassemble.webapp.database.alchemy_connection_string()
        clear_invalid_package_auth(start_time=start_time)
        if mode == 'initialize':
            logmessage("update: updating with mode initialize after " + str(time.time() - start_time) + " seconds")
            update_versions(start_time=start_time)
            any_package = db.session.execute(select(Package).filter_by(active=True)).first()
            if any_package is None:
                add_dependencies(1, start_time=start_time)
                update_versions(start_time=start_time)
            check_for_updates(start_time=start_time, invalidate_cache=False)
            if not SINGLE_SERVER:
                remove_inactive_hosts(start_time=start_time)
        else:
            logmessage("update: updating with mode check_for_updates after " + str(time.time() - start_time) + " seconds")
            check_for_updates(start_time=start_time)
            if USING_SUPERVISOR:
                SUPERVISORCTL = [daconfig.get('supervisorctl', 'supervisorctl')]
                if daconfig['supervisor'].get('username', None):
                    SUPERVISORCTL.extend(['--username', daconfig['supervisor']['username'], '--password', daconfig['supervisor']['password']])
                container_role = ':' + os.environ.get('CONTAINERROLE', '') + ':'
                if re.search(r':(web|celery|all):', container_role):
                    logmessage("update: sending reset signal after " + str(time.time() - start_time) + " seconds")
                    args = SUPERVISORCTL + ['-s', 'http://localhost:9001', 'start', 'reset']
                    subprocess.run(args, check=False)
                else:
                    logmessage("update: not sending reset signal because not web or celery after " + str(time.time() - start_time) + " seconds")
            else:
                logmessage("update: touched wsgi file after " + str(time.time() - start_time) + " seconds")
                wsgi_file = daconfig.get('webapp', '/usr/share/docassemble/webapp/docassemble.wsgi')
                if os.path.isfile(wsgi_file):
                    with open(wsgi_file, 'a', encoding='utf-8'):
                        os.utime(wsgi_file, None)
        db.engine.dispose()
    sys.exit(0)

if __name__ == "__main__":
    main()



"""add a column for the VoiceRSS voice

Revision ID: fd1547c94c46
Revises: a8746ddab8cb
Create Date: 2022-08-30 09:09:06.028272

"""
from alembic import op
import sqlalchemy as sa
from docassemble.webapp.database import dbtableprefix


# revision identifiers, used by Alembic.
revision = 'fd1547c94c46'
down_revision = 'a8746ddab8cb'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(dbtableprefix + 'speaklist', sa.Column('voice', sa.String(20)))


def downgrade():
    op.drop_column(dbtableprefix + 'speaklist', 'voice')



import os
import subprocess
import time
from packaging import version

from flask import Flask, abort, make_response, render_template, request
app = Flask(__name__)

LOG_DIRECTORY = '/var/www/html/log'

version_file = '/usr/share/docassemble/webapp/VERSION'
if os.path.isfile(version_file) and os.access(version_file, os.R_OK):
    with open(version_file, 'r', encoding='utf-8') as fp:
        system_version = fp.read().strip()
else:
    system_version = '0.1.12'

if version.parse(system_version) < version.parse('1.4.0'):
    READY_FILE = '/usr/share/docassemble/webapp/ready'
else:
    READY_FILE = '/var/run/docassemble/ready'


@app.route('/listlog')
def list_log_files():
    cmd = "supervisorctl "
    if os.getenv('DASUPERVISORUSERNAME', None):
        cmd += '--username ' + os.getenv('DASUPERVISORUSERNAME') + ' --password ' + os.getenv('DASUPERVISORPASSWORD') + ' '
    cmd += "--serverurl http://localhost:9001 start sync > /dev/null && while supervisorctl --serverurl http://localhost:9001 status sync | grep -q RUNNING; do sleep 1; done"
    result = subprocess.run(cmd, shell=True, check=False).returncode
    if result == 0:
        file_listing = [f for f in os.listdir(LOG_DIRECTORY) if os.path.isfile(os.path.join(LOG_DIRECTORY, f))]
        if len(file_listing) == 0:
            time.sleep(2)
            file_listing = [f for f in os.listdir(LOG_DIRECTORY) if os.path.isfile(os.path.join(LOG_DIRECTORY, f))]
        return "\n".join(sorted(file_listing))
    return "There was an error."


@app.route("/listlog/health_check", methods=['GET'])
def health_check():
    if request.args.get('ready', False):
        if not os.path.isfile(READY_FILE):
            abort(400)
    response = make_response(render_template('pages/health_check.html', content="OK"), 200)
    response.headers['Cache-Control'] = 'no-store, no-cache, must-revalidate, post-check=0, pre-check=0, max-age=0'
    return response



"""add unique constraints

Revision ID: a8746ddab8cb
Revises: 693d02299f38
Create Date: 2022-04-07 17:49:47.832279

"""
from alembic import op
from docassemble.webapp.database import dbtableprefix


# revision identifiers, used by Alembic.
revision = 'a8746ddab8cb'
down_revision = '693d02299f38'
branch_labels = None
depends_on = None


def upgrade():
    op.create_unique_constraint(dbtableprefix + 'chatlog_id_key', dbtableprefix + 'chatlog', ['id'])
    op.create_unique_constraint(dbtableprefix + 'email_id_key', dbtableprefix + 'email', ['id'])
    op.create_unique_constraint(dbtableprefix + 'emailattachment_id_key', dbtableprefix + 'emailattachment', ['id'])
    op.create_unique_constraint(dbtableprefix + 'globalobjectstorage_id_key', dbtableprefix + 'globalobjectstorage', ['id'])
    op.create_unique_constraint(dbtableprefix + 'jsonstorage_id_key', dbtableprefix + 'jsonstorage', ['id'])
    op.create_unique_constraint(dbtableprefix + 'machinelearning_id_key', dbtableprefix + 'machinelearning', ['id'])
    op.create_unique_constraint(dbtableprefix + 'objectstorage_id_key', dbtableprefix + 'objectstorage', ['id'])
    op.create_unique_constraint(dbtableprefix + 'role_id_key', dbtableprefix + 'role', ['id'])
    op.create_unique_constraint(dbtableprefix + 'shortener_id_key', dbtableprefix + 'shortener', ['id'])
    op.create_unique_constraint(dbtableprefix + 'speaklist_id_key', dbtableprefix + 'speaklist', ['id'])
    op.create_unique_constraint(dbtableprefix + 'supervisors_id_key', dbtableprefix + 'supervisors', ['id'])
    op.create_unique_constraint(dbtableprefix + 'tempuser_id_key', dbtableprefix + 'tempuser', ['id'])
    op.create_unique_constraint(dbtableprefix + 'uploads_indexno_key', dbtableprefix + 'uploads', ['indexno'])
    op.create_unique_constraint(dbtableprefix + 'uploadsroleauth_id_key', dbtableprefix + 'uploadsroleauth', ['id'])
    op.create_unique_constraint(dbtableprefix + 'uploadsuserauth_id_key', dbtableprefix + 'uploadsuserauth', ['id'])
    op.create_unique_constraint(dbtableprefix + 'user_id_key', dbtableprefix + 'user', ['id'])
    op.create_unique_constraint(dbtableprefix + 'user_auth_id_key', dbtableprefix + 'user_auth', ['id'])
    op.create_unique_constraint(dbtableprefix + 'user_invite_id_key', dbtableprefix + 'user_invite', ['id'])
    op.create_unique_constraint(dbtableprefix + 'user_roles_id_key', dbtableprefix + 'user_roles', ['id'])


def downgrade():
    op.drop_constraint(dbtableprefix + 'chatlog_id_key', dbtableprefix + 'chatlog')
    op.drop_constraint(dbtableprefix + 'email_id_key', dbtableprefix + 'email')
    op.drop_constraint(dbtableprefix + 'emailattachment_id_key', dbtableprefix + 'emailattachment')
    op.drop_constraint(dbtableprefix + 'globalobjectstorage_id_key', dbtableprefix + 'globalobjectstorage')
    op.drop_constraint(dbtableprefix + 'jsonstorage_id_key', dbtableprefix + 'jsonstorage')
    op.drop_constraint(dbtableprefix + 'machinelearning_id_key', dbtableprefix + 'machinelearning')
    op.drop_constraint(dbtableprefix + 'objectstorage_id_key', dbtableprefix + 'objectstorage')
    op.drop_constraint(dbtableprefix + 'role_id_key', dbtableprefix + 'role')
    op.drop_constraint(dbtableprefix + 'shortener_id_key', dbtableprefix + 'shortener')
    op.drop_constraint(dbtableprefix + 'speaklist_id_key', dbtableprefix + 'speaklist')
    op.drop_constraint(dbtableprefix + 'supervisors_id_key', dbtableprefix + 'supervisors')
    op.drop_constraint(dbtableprefix + 'tempuser_id_key', dbtableprefix + 'tempuser')
    op.drop_constraint(dbtableprefix + 'uploads_indexno_key', dbtableprefix + 'uploads')
    op.drop_constraint(dbtableprefix + 'uploadsroleauth_id_key', dbtableprefix + 'uploadsroleauth')
    op.drop_constraint(dbtableprefix + 'uploadsuserauth_id_key', dbtableprefix + 'uploadsuserauth')
    op.drop_constraint(dbtableprefix + 'user_id_key', dbtableprefix + 'user')
    op.drop_constraint(dbtableprefix + 'user_auth_id_key', dbtableprefix + 'user_auth')
    op.drop_constraint(dbtableprefix + 'user_invite_id_key', dbtableprefix + 'user_invite')
    op.drop_constraint(dbtableprefix + 'user_roles_id_key', dbtableprefix + 'user_roles')



import requests
import docassemble.base.util


def send_fax(fax_number, the_file, config, country=None):  # pylint: disable=unused-argument
    if not bool((hasattr(the_file, 'extension') and the_file.extension == 'pdf') or (hasattr(the_file, 'mimetype') and the_file.mimetype == 'application/pdf')):
        the_file = docassemble.base.util.pdf_concatenate(the_file)
    telnyx_api_endpoint = config.get('api endpoint', 'https://api.telnyx.com/v2/faxes')
    headers = {"Authorization": "Bearer " + config['api key']}
    media_url = the_file.url_for(temporary=True, seconds=600)
    r = requests.post(telnyx_api_endpoint, json={"media_url": media_url, "connection_id": config['app id'], "to": fax_number, "from": config['number']}, headers=headers)
    if r.status_code != 202:
        raise Exception("Exception when calling Telnyx: status code is " + str(r.status_code) + " and response is " + r.text)
    try:
        response = r.json()
    except:
        raise Exception("Exception when calling Telnyx: response not JSON: " + r.text)
    try:
        response['data']['id']
    except:
        raise Exception("Exception when calling Telnyx: id not in response: " + r.text)
    try:
        response['data']['latest_update_time'] = response['data']['created_at']
    except:
        raise Exception("Exception when calling Telnyx: could not find created_at time " + r.text)
    return response['data']



import json
import hashlib
import codecs
from docassemble.webapp.daredis import r


def add_specific_api_key(name, api_key, user_id, secret_key):
    info = dict(name=name, method='none', constraints=[])
    if not (isinstance(api_key, str) and len(api_key) == 32):
        return False
    info['last_four'] = api_key[-4:]
    new_api_key = encrypt_api_key(api_key, secret_key)
    if len(r.keys('da:apikey:userid:*:key:' + new_api_key + ':info')) > 0:
        return False
    for rkey in r.keys('da:apikey:userid:' + str(user_id) + ':key:*:info'):
        rkey = rkey.decode()
        try:
            info = json.loads(r.get(rkey).decode())
            assert isinstance(info, dict)
        except:
            continue
        if info['name'] == name:
            return False
    r.set('da:apikey:userid:' + str(user_id) + ':key:' + new_api_key + ':info', json.dumps(info))
    return True


def encrypt_api_key(key, secret_key):
    return codecs.encode(hashlib.pbkdf2_hmac('sha256', bytearray(key, 'utf-8'), bytearray(secret_key, encoding='utf-8'), 100000), 'base64').decode().strip()



import datetime
import sys
import os
import redis
import docassemble.base.config
if __name__ == "__main__":
    docassemble.base.config.load(arguments=sys.argv)
from docassemble.base.config import daconfig, parse_redis_uri
from docassemble.webapp.cloud import get_cloud
from docassemble.base.logger import logmessage


def errlog(text):
    logmessage(str(datetime.datetime.now()) + " " + text)


def main():
    container_role = ':' + os.environ.get('CONTAINERROLE', '') + ':'
    errlog("checking to see if running create_tables if necessary")
    if ':all:' in container_role or ':cron:' in container_role:
        (redis_host, redis_port, redis_username, redis_password, redis_offset, redis_cli, ssl_opts) = parse_redis_uri()  # pylint: disable=unused-variable
        r = redis.StrictRedis(host=redis_host, port=redis_port, db=redis_offset, password=redis_password, username=redis_username, **ssl_opts)
        if daconfig.get('ip address ban enabled', True):
            keys_to_delete = r.keys('da:failedlogin:ip:*')
            for key_to_delete in keys_to_delete:
                try:
                    r.delete(key_to_delete.decode())
                except:
                    pass
        if r.get('da:skip_create_tables'):
            logmessage("restart: skipping create_tables")
            r.delete('da:skip_create_tables')
        else:
            errlog("running create_tables")
            import docassemble.webapp.create_tables  # pylint: disable=redefined-outer-name,import-outside-toplevel
            docassemble.webapp.create_tables.main()
            errlog("finished create_tables")
        if ':cron:' in container_role:
            r.delete('da:cron_restart')

    webapp_path = daconfig.get('webapp', '/usr/share/docassemble/webapp/docassemble.wsgi')
    cloud = get_cloud()
    if cloud is not None:
        key = cloud.get_key('config.yml')
        if key.does_exist:
            key.get_contents_to_filename(daconfig['config file'])
            logmessage("Wrote config file based on copy on cloud")
    wsgi_file = webapp_path
    if os.path.isfile(wsgi_file):
        errlog("touching wsgi file")
        with open(wsgi_file, 'a', encoding='utf-8'):
            os.utime(wsgi_file, None)
        errlog("Restarted.")
    sys.exit(0)

if __name__ == "__main__":
    main()



from urllib.request import urlopen, Request
import importlib
import mimetypes
import os
import re
import subprocess
# import sys
import tempfile
import xml.etree.ElementTree as ET
import PyPDF2
from PIL import Image
from docassemble.base.generate_key import random_lower_string
from docassemble.base.logger import logmessage
import docassemble.base.config
import docassemble.base.functions
from docassemble.webapp.core.models import Uploads, UploadsUserAuth, UploadsRoleAuth
from docassemble.webapp.db_object import db
from docassemble.webapp.files import SavedFile, get_ext_and_mimetype
from docassemble.webapp.users.models import UserDictKeys, UserRoles
import docassemble.webapp.cloud
from flask import session, url_for
from flask_login import current_user
from sqlalchemy import and_, select

cloud = docassemble.webapp.cloud.get_cloud()

QPDF_PATH = 'qpdf'


def url_if_exists(file_reference, **kwargs):
    attach_parameter = '&attachment=1' if kwargs.get('_attachment', False) else ''
    parts = file_reference.split(":")
    base_url = url_for('rootindex', _external=kwargs.get('_external', False)).rstrip('/')
    if len(parts) == 2:
        if cloud and docassemble.base.config.daconfig.get('use cloud urls', False):
            m = re.search(r'^docassemble.playground([0-9]+)(.*)$', parts[0])
            if m:
                user_id = m.group(1)
                project = m.group(2)
                if re.search(r'^data/sources/', parts[1]):
                    section = 'playgroundsources'
                    filename = re.sub(r'^data/sources/', '', parts[1])
                    filename = re.sub(r'[^A-Za-z0-9\-\_\. ]', '', filename)
                    if project == '':
                        key = str(section) + '/' + str(user_id) + '/' + filename
                    else:
                        key = str(section) + '/' + str(user_id) + '/' + project + '/' + filename
                    cloud_key = cloud.get_key(key)
                    if cloud_key.does_exist:
                        if kwargs.get('_attachment', False):
                            return cloud_key.generate_url(3600, display_filename=filename)
                        return cloud_key.generate_url(3600)
                    return None
                section = 'playgroundstatic'
                filename = re.sub(r'^data/static/', '', parts[1])
                version_parameter = get_version_parameter(parts[0])
                return base_url + '/packagestatic/' + parts[0] + '/' + re.sub(r'^data/static/', '', parts[1]) + version_parameter + attach_parameter
        the_path = docassemble.base.functions.static_filename_path(file_reference)
        if the_path is None or not os.path.isfile(the_path):
            return None
        version_parameter = get_version_parameter(parts[0])
        return base_url + '/packagestatic/' + parts[0] + '/' + re.sub(r'^data/static/', '', parts[1]) + version_parameter + attach_parameter
    return None


def get_version_parameter(package):
    try:
        return '?v=' + str(importlib.import_module(package).__version__)
    except:
        if package.startswith('docassemble.playground'):
            return '?v=' + random_lower_string(6)
        return ''


def get_info_from_file_reference(file_reference, **kwargs):
    # logmessage('file reference is ' + str(file_reference))
    convert = kwargs.get('convert', None)
    privileged = kwargs.get('privileged', None)
    has_info = False
    if re.search(r'^[0-9]+$', str(file_reference)):
        uids = kwargs.get('uids', None)
        if uids is None or len(uids) == 0:
            new_uid = docassemble.base.functions.get_uid()
            if new_uid is not None:
                uids = [new_uid]
            else:
                uids = []
        if 'filename' in kwargs:
            result = get_info_from_file_number(int(file_reference), privileged=privileged, filename=kwargs['filename'], uids=uids)
        else:
            result = get_info_from_file_number(int(file_reference), privileged=privileged, uids=uids)
        if 'fullpath' not in result:
            result['fullpath'] = None
        has_info = True
    elif re.search(r'^https?://', str(file_reference)):
        # logmessage("get_info_from_file_reference: " + str(file_reference) + " is a URL")
        possible_filename = re.sub(r'.*/', '', file_reference)
        if possible_filename == '':
            possible_filename = 'index.html'
        if re.search(r'\.', possible_filename):
            (possible_ext, possible_mimetype) = get_ext_and_mimetype(possible_filename)
            possible_ext = re.sub(r'[^A-Za-z0-9\.].*', '', possible_ext)
            # logmessage("get_info_from_file_reference: starting with " + str(possible_ext) + " and " + str(possible_mimetype))
        else:
            possible_ext = 'txt'
            possible_mimetype = 'text/plain'
        result = {}
        temp_file = tempfile.NamedTemporaryFile(prefix="datemp", suffix='.' + possible_ext, delete=False)
        req = Request(file_reference, headers={'User-Agent': docassemble.base.config.daconfig.get('user agent', 'curl/7.64.0')})
        response = urlopen(req)
        temp_file.write(response.read())
        # (local_filename, headers) = urllib.urlretrieve(file_reference)
        result['fullpath'] = temp_file.name
        try:
            # result['mimetype'] = headers.gettype()
            result['mimetype'] = response.headers['Content-Type']
            # logmessage("get_info_from_file_reference: mimetype is " + str(result['mimetype']))
        except Exception:
            logmessage("get_info_from_file_reference: could not get mimetype from headers")
            result['mimetype'] = possible_mimetype
            result['extension'] = possible_ext
        if 'extension' not in result:
            # logmessage("get_info_from_file_reference: extension not in result")
            result['extension'] = re.sub(r'^\.', '', mimetypes.guess_extension(result['mimetype']))
            # logmessage("get_info_from_file_reference: extension is " + str(result['extension']))
        if re.search(r'\.', possible_filename):
            result['filename'] = possible_filename
        else:
            result['filename'] = possible_filename + '.' + result['extension']
        path_parts = os.path.splitext(result['fullpath'])
        result['path'] = path_parts[0]
        has_info = True
        # logmessage("get_info_from_file_reference: downloaded to " + str(result['fullpath']))
    else:
        # logmessage(str(file_reference) + " is not a URL")
        result = {}
        question = kwargs.get('question', None)
        manual_package = kwargs.get('package', None)
        folder = kwargs.get('folder', None)
        the_package = None
        parts = file_reference.split(':')
        if len(parts) == 1:
            the_package = None
            if question is not None:
                the_package = question.from_source.package
            elif manual_package is not None:
                the_package = manual_package
            if the_package is None:
                the_package = docassemble.base.functions.get_current_package()
            if folder is None:
                m = re.search(r'^data/(templates|sources|static)/(.*)', file_reference)
                if m:
                    folder = m.group(1)
                    file_reference = m.group(2)
            if folder is not None and not re.search(r'/', file_reference):
                file_reference = 'data/' + str(folder) + '/' + file_reference
            if the_package is not None:
                # logmessage("package is " + str(the_package))
                file_reference = the_package + ':' + file_reference
            else:
                # logmessage("package was null")
                file_reference = 'docassemble.base:' + file_reference
            if the_package is not None:
                result['package'] = the_package
        elif len(parts) == 2:
            result['package'] = parts[0]
        result['fullpath'] = docassemble.base.functions.static_filename_path(file_reference)
    # logmessage("path is " + str(result['fullpath']))
    if result['fullpath'] is not None:  # os.path.isfile(result['fullpath'])
        if not has_info:
            result['filename'] = os.path.basename(result['fullpath'])
            ext_type, result['mimetype'] = get_ext_and_mimetype(result['fullpath'])  # pylint: disable=unused-variable
            path_parts = os.path.splitext(result['fullpath'])
            result['path'] = path_parts[0]
            result['extension'] = path_parts[1].lower()
            result['extension'] = re.sub(r'\.', '', result['extension'])
        # logmessage("Extension is " + result['extension'])
        if convert is not None and result['extension'] in convert:
            # logmessage("Converting...")
            if os.path.isfile(result['path'] + '.' + convert[result['extension']]):
                # logmessage("Found conversion file ")
                result['extension'] = convert[result['extension']]
                result['fullpath'] = result['path'] + '.' + result['extension']
                ext_type, result['mimetype'] = get_ext_and_mimetype(result['fullpath'])
            else:
                logmessage("Did not find file " + result['path'] + '.' + convert[result['extension']])
                return {}
        # logmessage("Full path is " + result['fullpath'])
        if os.path.isfile(result['fullpath']) and not has_info:
            add_info_about_file(result['fullpath'], result['path'], result)
    else:
        logmessage("File reference " + str(file_reference) + " DID NOT EXIST.")
    return result


def safe_pypdf_reader(filename):
    try:
        return PyPDF2.PdfFileReader(open(filename, 'rb'), overwriteWarnings=False)
    except PyPDF2.utils.PdfReadError:
        new_filename = tempfile.NamedTemporaryFile(prefix="datemp", mode="wb", suffix=".pdf", delete=False)
        qpdf_subprocess_arguments = [QPDF_PATH, filename, new_filename.name]
        try:
            result = subprocess.run(qpdf_subprocess_arguments, timeout=60, check=False).returncode
        except subprocess.TimeoutExpired:
            result = 1
        if result != 0:
            raise Exception("Call to qpdf failed for template " + str(filename) + " where arguments were " + " ".join(qpdf_subprocess_arguments))
        return PyPDF2.PdfFileReader(open(new_filename.name, 'rb'), overwriteWarnings=False)


def add_info_about_file(filename, basename, result):
    if result['extension'] == 'pdf':
        try:
            reader = safe_pypdf_reader(filename)
            result['encrypted'] = reader.isEncrypted
            try:
                if '/AcroForm' in reader.trailer['/Root']:
                    result['acroform'] = True
                else:
                    result['acroform'] = False
            except:
                result['acroform'] = False
            result['pages'] = reader.getNumPages()
        except:
            result['pages'] = 1
    elif os.path.isfile(basename + '.pdf'):
        try:
            reader = safe_pypdf_reader(basename + '.pdf')
            result['encrypted'] = reader.isEncrypted
            try:
                if '/AcroForm' in reader.trailer['/Root']:
                    result['acroform'] = True
                else:
                    result['acroform'] = False
            except:
                result['acroform'] = False
            result['pages'] = reader.getNumPages()
        except:
            result['pages'] = 1
    elif result['extension'] in ['png', 'jpg', 'gif']:
        im = Image.open(filename)
        result['width'], result['height'] = im.size
    elif result['extension'] == 'svg':
        try:
            tree = ET.parse(filename)
            root = tree.getroot()
            viewBox = root.attrib.get('viewBox', None)
            if viewBox is not None:
                dimen = viewBox.split(' ')
                if len(dimen) == 4:
                    result['width'] = float(dimen[2]) - float(dimen[0])
                    result['height'] = float(dimen[3]) - float(dimen[1])
        except:
            raise Exception("problem reading " + str(filename))
            # logmessage('add_info_about_file: could not read ' + str(filename))


def get_info_from_file_number(file_number, privileged=False, filename=None, uids=None):
    if current_user and current_user.is_authenticated and current_user.has_role('admin', 'developer', 'advocate', 'trainer'):
        privileged = True
    elif uids is None or len(uids) == 0:
        new_uid = docassemble.base.functions.get_uid()
        if new_uid is not None:
            uids = [new_uid]
        else:
            uids = []
    result = {}
    upload = db.session.execute(select(Uploads).filter_by(indexno=file_number)).scalar()
    if not privileged and upload is not None and upload.private and upload.key not in uids:
        has_access = False
        if current_user and current_user.is_authenticated:
            if db.session.execute(select(UserDictKeys).filter_by(key=upload.key, user_id=current_user.id)).first() or db.session.execute(select(UploadsUserAuth).filter_by(uploads_indexno=file_number, user_id=current_user.id)).first() or db.session.execute(select(UploadsRoleAuth.id).join(UserRoles, and_(UserRoles.user_id == current_user.id, UploadsRoleAuth.role_id == UserRoles.role_id)).where(UploadsRoleAuth.uploads_indexno == file_number)).first():
                has_access = True
        elif session and 'tempuser' in session:
            temp_user_id = int(session['tempuser'])
            if db.session.execute(select(UserDictKeys).filter_by(key=upload.key, temp_user_id=temp_user_id)).first() or db.session.execute(select(UploadsUserAuth).filter_by(uploads_indexno=file_number, temp_user_id=temp_user_id)).first():
                has_access = True
        if not has_access:
            upload = None
    if upload:
        if filename is None:
            result['filename'] = upload.filename
        else:
            result['filename'] = filename
        result['extension'], result['mimetype'] = get_ext_and_mimetype(result['filename'])
        sf = SavedFile(file_number, extension=result['extension'], fix=True)
        result['path'] = sf.path
        result['fullpath'] = result['path'] + '.' + result['extension']
        result['private'] = upload.private
        result['persistent'] = upload.persistent
        # logmessage("fullpath is " + str(result['fullpath']))
    if 'path' not in result:
        logmessage("get_info_from_file_number: path is not in result for " + str(file_number))
        return result
    final_filename = result['path'] + '.' + result['extension']
    if os.path.isfile(final_filename):
        add_info_about_file(final_filename, result['path'], result)
    # else:
    #     logmessage("Filename " + final_filename + "did not exist.")
    return result



import redis
from docassemble.base.config import parse_redis_uri

(redis_host, redis_port, redis_username, redis_password, redis_offset, redis_cli, ssl_opts) = parse_redis_uri()

r = redis.StrictRedis(host=redis_host, port=redis_port, db=redis_offset, password=redis_password, username=redis_username, **ssl_opts)
r_store = redis.StrictRedis(host=redis_host, port=redis_port, db=1 + redis_offset, password=redis_password, username=redis_username, **ssl_opts)
r_user = redis.StrictRedis(host=redis_host, port=redis_port, db=2 + redis_offset, password=redis_password, username=redis_username, **ssl_opts)



import os
import sys
import docassemble.base.config
if __name__ == "__main__":
    docassemble.base.config.load(arguments=sys.argv)
from docassemble.base.config import daconfig
from docassemble.base.logger import logmessage


def main():
    webapp_path = daconfig.get('webapp', '/usr/share/docassemble/webapp/docassemble.wsgi')
    wsgi_file = webapp_path
    if os.path.isfile(wsgi_file):
        with open(wsgi_file, 'a', encoding='utf-8'):
            os.utime(wsgi_file, None)
            logmessage("Restarted WSGI.\n")
    sys.exit(0)

if __name__ == "__main__":
    main()



import sys
import os
import time
import datetime
import docassemble.base.config
if __name__ == "__main__":
    docassemble.base.config.load(arguments=sys.argv)
from docassemble.base.config import daconfig
from docassemble.base.functions import word
from docassemble.base.generate_key import random_alphanumeric
from docassemble.base.logger import logmessage
from docassemble.webapp.app_object import app
from docassemble.webapp.core.models import Uploads, ObjectStorage, SpeakList, Shortener, MachineLearning, GlobalObjectStorage
from docassemble.webapp.database import alchemy_connection_string, dbtableprefix
from docassemble.webapp.db_object import db
from docassemble.webapp.packages.models import Package
from docassemble.webapp.update import add_dependencies
from docassemble.webapp.users.models import UserModel, UserAuthModel, Role, UserDict, UserDictKeys, ChatLog, TempUser, MyUserInvitation, UserRoles
from docassemble.webapp.database import dbprefix
import docassemble.webapp.core.models
import docassemble.webapp.packages.models
from docassemble.webapp.api_key import add_specific_api_key
from sqlalchemy import select, delete, inspect
from sqlalchemy.sql import text
# import random
# import string
from docassemble_flask_user import UserManager, SQLAlchemyAdapter
import pkg_resources
from alembic.config import Config
from alembic import command


def get_role(the_db, name, result=None):
    if result is None:
        result = {}
    the_role = the_db.session.execute(select(Role).filter_by(name=name)).first()
    if the_role:
        return the_role
    the_role = Role(name=name)
    the_db.session.add(the_role)
    the_db.session.commit()
    result['changed'] = True
    return the_role


def get_user(the_db, role, defaults, result=None):
    if result is None:
        result = {}
    the_user = the_db.session.execute(select(UserModel).filter_by(nickname=defaults['nickname'])).scalar()
    if the_user:
        return the_user
    while True:
        new_social = 'local$' + random_alphanumeric(32)
        existing_user = the_db.session.execute(select(UserModel).filter_by(social_id=new_social)).scalar()
        if existing_user:
            continue
        break
    user_auth = UserAuthModel(password=app.user_manager.hash_password(defaults.get('password', 'password')))
    the_user = UserModel(
        active=defaults.get('active', True),
        nickname=defaults['nickname'],
        social_id=new_social,
        email=defaults['email'],
        user_auth=user_auth,
        first_name=defaults.get('first_name', ''),
        last_name=defaults.get('last_name', ''),
        country=defaults.get('country', ''),
        subdivisionfirst=defaults.get('subdivisionfirst', ''),
        subdivisionsecond=defaults.get('subdivisionsecond', ''),
        subdivisionthird=defaults.get('subdivisionthird', ''),
        organization=defaults.get('organization', ''),
        confirmed_at=datetime.datetime.now()
    )
    the_user.roles.append(role)
    the_db.session.add(user_auth)
    the_db.session.add(the_user)
    the_db.session.commit()
    result['changed'] = True
    return the_user


def test_for_errors(start_time=None):
    todo = [['chatlog', 'id', ChatLog],
            ['email', 'id', docassemble.webapp.core.models.Email],
            ['emailattachment', 'id', docassemble.webapp.core.models.EmailAttachment],
            ['globalobjectstorage', 'id', docassemble.webapp.core.models.GlobalObjectStorage],
            ['install', 'id', docassemble.webapp.packages.models.Install],
            ['jsonstorage', 'id', docassemble.webapp.core.models.JsonStorage],
            ['machinelearning', 'id', docassemble.webapp.core.models.MachineLearning],
            ['objectstorage', 'id', docassemble.webapp.core.models.ObjectStorage],
            ['package_auth', 'id', docassemble.webapp.packages.models.PackageAuth],
            ['package', 'id', Package],
            ['role', 'id', Role],
            ['shortener', 'id', docassemble.webapp.core.models.Shortener],
            ['speaklist', 'id', docassemble.webapp.core.models.SpeakList],
            ['supervisors', 'id', docassemble.webapp.core.models.Supervisors],
            ['tempuser', 'id', TempUser],
            ['uploads', 'indexno', docassemble.webapp.core.models.Uploads],
            ['uploadsroleauth', 'id', docassemble.webapp.core.models.UploadsRoleAuth],
            ['uploadsuserauth', 'id', docassemble.webapp.core.models.UploadsUserAuth],
            ['user_auth', 'id', UserAuthModel],
            ['user', 'id', UserModel],
            ['user_invite', 'id', MyUserInvitation],
            ['user_roles', 'id', UserRoles],
            ['userdict', 'indexno', UserDict],
            ['userdictkeys', 'indexno', UserDictKeys]]
    for table, column, tableclass in todo:
        last_value = 0
        for results in db.session.execute(text("select last_value from " + table + "_" + column + "_seq")):
            last_value = results[0]
        max_value = db.session.execute(select(db.func.max(getattr(tableclass, column)))).scalar()
        if max_value is not None and max_value > last_value:
            logmessage('create_tables.test_for_errors: ' + table + " has an error: " + str(last_value) + " " + str(max_value) + " after " + str(time.time() - start_time) + " seconds.")
            db.session.execute(text("alter sequence " + table + "_" + column + "_seq restart with :newval"), {'newval': max_value + 1})
            db.session.commit()


def populate_tables(start_time=None):
    if start_time is None:
        start_time = time.time()
    logmessage("create_tables.populate_tables: starting after " + str(time.time() - start_time) + " seconds.")
    UserManager(SQLAlchemyAdapter(db, UserModel, UserAuthClass=UserAuthModel), app)
    logmessage("create_tables.populate_tables: obtained UserManager after " + str(time.time() - start_time) + " seconds.")
    result = {}
    admin_defaults = daconfig.get('default admin account', {})
    if 'email' not in admin_defaults:
        admin_defaults['email'] = os.getenv('DA_ADMIN_EMAIL', 'admin@admin.com')
    if 'nickname' not in admin_defaults:
        admin_defaults['nickname'] = 'admin'
    if 'first_name' not in admin_defaults:
        admin_defaults['first_name'] = word('System')
    if 'last_name' not in admin_defaults:
        admin_defaults['last_name'] = word('Administrator')
    if 'password' not in admin_defaults:
        admin_defaults['password'] = os.getenv('DA_ADMIN_PASSWORD', 'password')
    if 'api key' not in admin_defaults:
        api_key = os.getenv('DA_ADMIN_API_KEY', '')
        if api_key:
            admin_defaults['api key'] = api_key
    cron_defaults = daconfig.get('default cron account', {'nickname': 'cron', 'email': 'cron@admin.com', 'first_name': 'Cron', 'last_name': 'User'})
    cron_defaults['active'] = False
    user_role = get_role(db, 'user', result=result)
    admin_role = get_role(db, 'admin', result=result)
    cron_role = get_role(db, 'cron', result=result)
    get_role(db, 'customer', result=result)
    get_role(db, 'developer', result=result)
    get_role(db, 'advocate', result=result)
    get_role(db, 'trainer', result=result)
    if daconfig.get('fix user roles', False):
        logmessage("create_tables.populate_tables: fixing user roles after " + str(time.time() - start_time) + " seconds.")
        to_fix = []
        for user in db.session.execute(select(UserModel).options(db.joinedload(UserModel.roles))).scalars():
            if len(user.roles) == 0:
                to_fix.append(user)
        if len(to_fix) > 0:
            logmessage("create_tables.populate_tables: found user roles to fix after " + str(time.time() - start_time) + " seconds.")
            for user in to_fix:
                user.roles.append(user_role)
            db.session.commit()
        logmessage("create_tables.populate_tables: done fixing user roles after " + str(time.time() - start_time) + " seconds.")
    admin = get_user(db, admin_role, admin_defaults, result=result)
    cron = get_user(db, cron_role, cron_defaults, result=result)
    if admin.confirmed_at is None:
        admin.confirmed_at = datetime.datetime.now()
    if cron.confirmed_at is None:
        cron.confirmed_at = datetime.datetime.now()
    if result.get('changed', False):
        db.session.commit()
    if 'api key' in admin_defaults:
        api_result = add_specific_api_key('default', admin_defaults['api key'], admin.id, daconfig.get('secretkey', '38ihfiFehfoU34mcq_4clirglw3g4o87'))
        if api_result:
            logmessage("create_tables.populate_tables: added API key")
    logmessage("create_tables.populate_tables: calling add_dependencies after " + str(time.time() - start_time) + " seconds.")
    add_dependencies(admin.id, start_time=start_time)
    logmessage("create_tables.populate_tables: add_dependencies finished after " + str(time.time() - start_time) + " seconds.")
    git_packages = db.session.execute(select(Package).filter_by(type='git')).scalars().all()
    package_info_changed = False
    for package in git_packages:
        if package.name in ['docassemble', 'docassemble.base', 'docassemble.webapp', 'docassemble.demo']:
            if package.giturl:
                package.giturl = None
                package_info_changed = True
            if package.gitsubdir:
                package.gitsubdir = None
                package_info_changed = True
            if package.type != 'pip':
                package.type = 'pip'
                package_info_changed = True
            if daconfig.get('stable version', False):
                if package.limitation != '<1.1.0':
                    package.limitation = '<1.1.0'
                    package_info_changed = True
    if package_info_changed:
        db.session.commit()
    logmessage("create_tables.populate_tables: ending after " + str(time.time() - start_time) + " seconds.")


def main():
    logmessage("create_tables.main: starting")
    start_time = time.time()
    if dbprefix.startswith('postgresql') and not daconfig.get('force text to varchar upgrade', False):
        do_varchar_upgrade = False
    else:
        do_varchar_upgrade = True
    with app.app_context():
        logmessage("create_tables.main: inside app context after " + str(time.time() - start_time) + " seconds.")
        if daconfig.get('use alembic', True):
            logmessage("create_tables.main: running alembic after " + str(time.time() - start_time) + " seconds.")
            insp = inspect(db.engine)
            if do_varchar_upgrade:
                changed = False
                if insp.has_table(dbtableprefix + 'userdict'):
                    db.session.execute(delete(UserDict).where(db.func.length(UserDict.filename) > 255).execution_options(synchronize_session=False))
                    changed = True
                if insp.has_table(dbtableprefix + 'userdictkeys'):
                    db.session.execute(delete(UserDictKeys).where(db.func.length(UserDictKeys.filename) > 255).execution_options(synchronize_session=False))
                    changed = True
                if insp.has_table(dbtableprefix + 'chatlog'):
                    db.session.execute(delete(ChatLog).where(db.func.length(ChatLog.filename) > 255).execution_options(synchronize_session=False))
                    changed = True
                if insp.has_table(dbtableprefix + 'uploads'):
                    db.session.execute(delete(Uploads).where(db.func.length(Uploads.filename) > 255).execution_options(synchronize_session=False))
                    db.session.execute(delete(Uploads).where(db.func.length(Uploads.yamlfile) > 255).execution_options(synchronize_session=False))
                    changed = True
                if insp.has_table(dbtableprefix + 'objectstorage'):
                    db.session.execute(delete(ObjectStorage).where(db.func.length(ObjectStorage.key) > 1024).execution_options(synchronize_session=False))
                    changed = True
                if insp.has_table(dbtableprefix + 'speaklist'):
                    db.session.execute(delete(SpeakList).where(db.func.length(SpeakList.filename) > 255).execution_options(synchronize_session=False))
                    changed = True
                if insp.has_table(dbtableprefix + 'shortener'):
                    db.session.execute(delete(Shortener).where(db.func.length(Shortener.filename) > 255).execution_options(synchronize_session=False))
                    db.session.execute(delete(Shortener).where(db.func.length(Shortener.key) > 255).execution_options(synchronize_session=False))
                    changed = True
                if insp.has_table(dbtableprefix + 'machinelearning'):
                    db.session.execute(delete(MachineLearning).where(db.func.length(MachineLearning.key) > 1024).execution_options(synchronize_session=False))
                    db.session.execute(delete(MachineLearning).where(db.func.length(MachineLearning.group_id) > 1024).execution_options(synchronize_session=False))
                    changed = True
                if insp.has_table(dbtableprefix + 'globalobjectstorage'):
                    db.session.execute(delete(GlobalObjectStorage).where(db.func.length(GlobalObjectStorage.key) > 1024).execution_options(synchronize_session=False))
                    changed = True
                if changed:
                    db.session.commit()
            packagedir = pkg_resources.resource_filename(pkg_resources.Requirement.parse('docassemble.webapp'), 'docassemble/webapp')
            if not os.path.isdir(packagedir):
                sys.exit("path for running alembic could not be found")
            alembic_cfg = Config(os.path.join(packagedir, 'alembic.ini'))
            alembic_cfg.set_main_option("sqlalchemy.url", alchemy_connection_string())
            alembic_cfg.set_main_option("script_location", os.path.join(packagedir, 'alembic'))
            if not insp.has_table(dbtableprefix + 'alembic_version'):
                logmessage("create_tables.main: creating alembic stamp")
                command.stamp(alembic_cfg, "head")
                logmessage("create_tables.main: done creating alembic stamp after " + str(time.time() - start_time) + " seconds.")
            if insp.has_table(dbtableprefix + 'user'):
                logmessage("create_tables.main: creating alembic stamp")
                logmessage("create_tables.main: running alembic upgrade")
                command.upgrade(alembic_cfg, "head")
                logmessage("create_tables.main: done running alembic upgrade after " + str(time.time() - start_time) + " seconds.")
        # db.drop_all()
        try:
            logmessage("create_tables.main: trying to create tables")
            db.create_all()
        except:
            logmessage("create_tables.main: error trying to create tables; trying a second time.")
            try:
                db.create_all()
            except:
                logmessage("create_tables.main: error trying to create tables; trying a third time.")
                db.create_all()
        logmessage("create_tables.main: finished creating tables after " + str(time.time() - start_time) + " seconds.")
        if dbprefix.startswith('postgresql'):
            try:
                test_for_errors(start_time=start_time)
            except:
                logmessage("create_tables.main: unable to test for errors after " + str(time.time() - start_time) + " seconds.")
        logmessage("create_tables.main: populating tables after " + str(time.time() - start_time) + " seconds.")
        populate_tables(start_time=start_time)
        logmessage("create_tables.main: disposing engine after " + str(time.time() - start_time) + " seconds.")
        db.engine.dispose()
    logmessage("create_tables.main: finishing after " + str(time.time() - start_time) + " seconds.")

if __name__ == "__main__":
    main()



# import sys
# from werkzeug.contrib.fixers import ProxyFix
from flask import Flask
# from flask_sqlalchemy import SQLAlchemy
from flask_socketio import SocketIO
import docassemble.webapp.db_object  # noqa: F401 # pylint: disable=unused-import
import sqlalchemy
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, backref


def create_app():
    the_app = Flask(__name__)
    the_app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    from docassemble.base.config import daconfig  # pylint: disable=import-outside-toplevel
    import docassemble.webapp.database  # pylint: disable=import-outside-toplevel,redefined-outer-name
    alchemy_connect_string = docassemble.webapp.database.alchemy_connection_string()
    # the_app.config['SQLALCHEMY_DATABASE_URI'] = alchemy_connect_string
    the_app.secret_key = daconfig.get('secretkey', '38ihfiFehfoU34mcq_4clirglw3g4o87')
    # the_app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    # the_db = SQLAlchemy(the_app)
    if alchemy_connect_string.startswith('postgresql'):
        connect_args = docassemble.webapp.database.connect_args()
        the_db = sqlalchemy.create_engine(alchemy_connect_string, connect_args=connect_args, pool_pre_ping=docassemble.webapp.database.pool_pre_ping)
    else:
        the_db = sqlalchemy.create_engine(alchemy_connect_string, pool_pre_ping=docassemble.webapp.database.pool_pre_ping)
    Base = declarative_base()
    Base.metadata.bind = the_db
    # the_app.wsgi_app = ProxyFix(the_app.wsgi_app)
    the_db.Model = Base
    the_db.Column = sqlalchemy.Column
    the_db.Integer = sqlalchemy.Integer
    the_db.String = sqlalchemy.String
    the_db.Index = sqlalchemy.Index
    the_db.Boolean = sqlalchemy.Boolean
    the_db.Text = sqlalchemy.Text
    the_db.DateTime = sqlalchemy.DateTime
    the_db.func = sqlalchemy.func
    the_db.relationship = relationship
    the_db.backref = backref
    the_db.ForeignKey = sqlalchemy.ForeignKey
    docassemble.webapp.db_object.db = the_db
    # import flask_login
    docassemble.webapp.db_object.UserMixin = object
    if 'cross site domains' in daconfig and isinstance(daconfig['cross site domains'], list) and len(daconfig['cross site domains']) > 0:
        origins = daconfig['cross site domains']
    else:
        origins = [daconfig.get('url root', '*')]
    the_socketio = SocketIO(the_app, async_mode='eventlet', verify=False, logger=True, engineio_logger=True, cors_allowed_origins=origins)
    return the_app, the_db, the_socketio

app, db, socketio = create_app()
